# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from apitools.base.py import exceptions
import datetime
import mock
import os
import tempfile
import unittest2
from expects import be_false, be_none, be_true, expect, equal, raise_error

from endpoints_management.control import (
    caches, check_request, client, quota_request, report_request, sc_messages
)


class TestSimpleLoader(unittest2.TestCase):
    SERVICE_NAME = u'simpler-loader'

    @mock.patch(u"endpoints_management.control.client.ReportOptions", autospec=True)
    @mock.patch(u"endpoints_management.control.client.QuotaOptions", autospec=True)
    @mock.patch(u"endpoints_management.control.client.CheckOptions", autospec=True)
    def test_should_create_client_ok(self, check_opts, quota_opts, report_opts):
        # the mocks return fake instances else code using them fails
        check_opts.return_value = caches.CheckOptions()
        report_opts.return_value = caches.ReportOptions()
        quota_opts.return_value = caches.QuotaOptions()

        # ensure the client is constructed using no args instances of the opts
        expect(client.Loaders.DEFAULT.load(self.SERVICE_NAME)).not_to(be_none)
        check_opts.assert_called_once_with()
        quota_opts.assert_called_once_with()
        report_opts.assert_called_once_with()

_TEST_CONFIG = u"""{
    "checkAggregatorConfig": {
       "cacheEntries": 10,
       "responseExpirationMs": 1000,
       "flushIntervalMs": 2000
    },
    "quotaAggregatorConfig": {
       "cacheEntries": 10,
       "expirationMs": 2000,
       "flushIntervalMs": 1000
    },
    "reportAggregatorConfig": {
       "cacheEntries": 10,
       "flushIntervalMs": 1000
    }
}
"""


class TestEnvironmentLoader(unittest2.TestCase):
    SERVICE_NAME = u'environment-loader'

    def setUp(self):
        json_fd = tempfile.NamedTemporaryFile(delete=False)
        with json_fd as f:
            f.write(_TEST_CONFIG.encode('ascii'))
        self._config_file = json_fd.name
        os.environ[client.CONFIG_VAR] = self._config_file

    def tearDown(self):
        if os.path.exists(self._config_file):
            os.remove(self._config_file)

    @mock.patch(u"endpoints_management.control.client.ReportOptions", autospec=True)
    @mock.patch(u"endpoints_management.control.client.QuotaOptions", autospec=True)
    @mock.patch(u"endpoints_management.control.client.CheckOptions", autospec=True)
    def test_should_create_client_from_environment_ok(self, check_opts, quota_opts, report_opts):
        check_opts.return_value = caches.CheckOptions()
        report_opts.return_value = caches.ReportOptions()
        quota_opts.return_value = caches.QuotaOptions()

        # ensure the client is constructed using options values from the test JSON
        expect(client.Loaders.ENVIRONMENT.load(self.SERVICE_NAME)).not_to(be_none)
        check_opts.assert_called_once_with(expiration=datetime.timedelta(0, 1),
                                           flush_interval=datetime.timedelta(0, 2),
                                           num_entries=10)
        report_opts.assert_called_once_with(flush_interval=datetime.timedelta(0, 1),
                                            num_entries=10)

    @mock.patch(u"endpoints_management.control.client.ReportOptions", autospec=True)
    @mock.patch(u"endpoints_management.control.client.QuotaOptions", autospec=True)
    @mock.patch(u"endpoints_management.control.client.CheckOptions", autospec=True)
    def test_should_use_defaults_if_file_is_missing(self, check_opts, quota_opts, report_opts):
        os.remove(self._config_file)
        self._assert_called_with_no_args_options(check_opts, quota_opts, report_opts)

    @mock.patch(u"endpoints_management.control.client.ReportOptions", autospec=True)
    @mock.patch(u"endpoints_management.control.client.QuotaOptions", autospec=True)
    @mock.patch(u"endpoints_management.control.client.CheckOptions", autospec=True)
    def test_should_use_defaults_if_file_is_missing(self, check_opts, quota_opts, report_opts):
        del os.environ[client.CONFIG_VAR]
        self._assert_called_with_no_args_options(check_opts, quota_opts, report_opts)

    @mock.patch(u"endpoints_management.control.client.ReportOptions")
    @mock.patch(u"endpoints_management.control.client.QuotaOptions")
    @mock.patch(u"endpoints_management.control.client.CheckOptions")
    def test_should_use_defaults_if_json_is_bad(self, check_opts, quota_opts, report_opts):
        with open(self._config_file, u'w') as f:
            f.write(_TEST_CONFIG + u'\n{ this will not parse as json}')
        self._assert_called_with_no_args_options(check_opts, quota_opts, report_opts)

    def _assert_called_with_no_args_options(self, check_opts, quota_opts, report_opts):
        # the mocks return fake instances else code using them fails
        check_opts.return_value = caches.CheckOptions()
        report_opts.return_value = caches.ReportOptions()
        quota_opts.return_value = caches.QuotaOptions()

        # ensure the client is constructed using no args instances of the opts
        expect(client.Loaders.ENVIRONMENT.load(self.SERVICE_NAME)).not_to(be_none)
        check_opts.assert_called_once_with()
        quota_opts.assert_called_once_with()
        report_opts.assert_called_once_with()


def _make_dummy_report_request(project_id, service_name):
    rules = report_request.ReportingRules()
    info = report_request.Info(
        consumer_project_id=project_id,
        operation_id=u'an_op_id',
        operation_name=u'an_op_name',
        method=u'GET',
        referer=u'a_referer',
        service_name=service_name)
    return info.as_report_request(rules)


def _make_dummy_quota_request(project_id, service_name):
    info = quota_request.Info(
        consumer_project_id=project_id,
        operation_id=u'an_op_id',
        operation_name=u'an_op_name',
        referer=u'a_referer',
        service_name=service_name,
        quota_info={'foo': 1, 'bar': 2})
    return info.as_allocate_quota_request()


def _make_dummy_check_request(project_id, service_name):
    info = check_request.Info(
        consumer_project_id=project_id,
        operation_id=u'an_op_id',
        operation_name=u'an_op_name',
        referer=u'a_referer',
        service_name=service_name)
    return info.as_check_request()


class TestClientStartAndStop(unittest2.TestCase):
    SERVICE_NAME = u'start-and-stop'
    PROJECT_ID = SERVICE_NAME + u'.project'

    def setUp(self):
        self._mock_transport = mock.MagicMock()
        self._subject = client.Loaders.DEFAULT.load(
            self.SERVICE_NAME,
            create_transport=lambda: self._mock_transport)

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    def test_should_create_a_thread_when_started(self, thread_class):
        self._subject.start()
        expect(thread_class.called).to(be_true)
        expect(len(thread_class.call_args_list)).to(equal(1))

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    def test_should_only_create_thread_on_first_start(self, thread_class):
        self._subject.start()
        self._subject.start()
        expect(len(thread_class.call_args_list)).to(equal(1))

    def test_should_noop_stop_if_not_started(self):
        # stop the subject, the transport should not see a request
        self._subject.stop()
        expect(self._mock_transport.called).to(be_false)

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    def test_should_clear_requests_on_stop(self, dummy_thread_class):
        # stop the subject, the transport did not see a request
        self._subject.start()
        self._subject.report(
            _make_dummy_report_request(self.PROJECT_ID, self.SERVICE_NAME))
        self._subject.stop()
        expect(self._mock_transport.services.Report.called).to(be_true)

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    def test_should_ignore_stop_if_already_stopped(self, dummy_thread_class):
        # stop the subject, the transport did not see a request
        self._subject.start()
        self._subject.report(
            _make_dummy_report_request(self.PROJECT_ID, self.SERVICE_NAME))
        self._subject.stop()
        self._mock_transport.reset_mock()
        self._subject.stop()
        expect(self._mock_transport.services.Report.called).to(be_false)

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    def test_should_ignore_bad_transport_when_not_cached(self, dummy_thread_class):
        self._subject.start()
        self._subject.report(
            _make_dummy_report_request(self.PROJECT_ID, self.SERVICE_NAME))
        self._mock_transport.services.Report.side_effect = exceptions.Error()
        self._subject.stop()
        expect(self._mock_transport.services.Report.called).to(be_true)


class TestClientCheck(unittest2.TestCase):
    SERVICE_NAME = u'check'
    PROJECT_ID = SERVICE_NAME + u'.project'

    def setUp(self):
        self._mock_transport = mock.MagicMock()
        self._subject = client.Loaders.DEFAULT.load(
            self.SERVICE_NAME,
            create_transport=lambda: self._mock_transport)

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    def test_should_start_itself_on_check_without_start(self, dummy_thread_class):
        dummy_request = _make_dummy_check_request(self.PROJECT_ID,
                                                  self.SERVICE_NAME)
        self._subject.check(dummy_request)
        assert self._subject._running

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    def test_should_send_the_request_if_not_cached(self, dummy_thread_class):
        self._subject.start()
        dummy_request = _make_dummy_check_request(self.PROJECT_ID,
                                                  self.SERVICE_NAME)
        self._subject.check(dummy_request)
        expect(self._mock_transport.services.Check.called).to(be_true)

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    def test_should_not_send_the_request_if_cached(self, dummy_thread_class):
        t = self._mock_transport
        self._subject.start()
        dummy_request = _make_dummy_check_request(self.PROJECT_ID,
                                                  self.SERVICE_NAME)
        dummy_response = sc_messages.CheckResponse(
            operationId=dummy_request.checkRequest.operation.operationId)
        t.services.Check.return_value = dummy_response
        expect(self._subject.check(dummy_request)).to(equal(dummy_response))
        t.reset_mock()
        expect(self._subject.check(dummy_request)).to(equal(dummy_response))
        expect(t.services.Check.called).to(be_false)

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    def test_should_return_null_if_transport_fails(self, dummy_thread_class):
        self._subject.start()
        dummy_request = _make_dummy_check_request(self.PROJECT_ID,
                                                  self.SERVICE_NAME)
        self._mock_transport.services.Check.side_effect = exceptions.Error()
        expect(self._subject.check(dummy_request)).to(be_none)


class TestClientQuota(unittest2.TestCase):
    SERVICE_NAME = u'quota'
    PROJECT_ID = SERVICE_NAME + u'.project'

    def setUp(self):
        self._mock_transport = mock.MagicMock()
        self._subject = client.Loaders.DEFAULT.load(
            self.SERVICE_NAME,
            create_transport=lambda: self._mock_transport)

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    def test_should_start_itself_on_quota_without_start(self, dummy_thread_class):
        dummy_request = _make_dummy_quota_request(self.PROJECT_ID,
                                                  self.SERVICE_NAME)
        self._subject.allocate_quota(dummy_request)
        assert self._subject._running

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    def test_should_queue_the_request_if_not_cached(self, dummy_thread_class):
        # the request isn't sent immediately as long as a cache exists
        # instead we get a temporary response
        self._subject.start()
        dummy_request = _make_dummy_quota_request(self.PROJECT_ID,
                                                  self.SERVICE_NAME)
        resp = self._subject.allocate_quota(dummy_request)
        with self._subject._quota_aggregator._out as out_deque:
            expect(out_deque[0]).to(equal(dummy_request))
        expect(resp.operationId).to(equal(
            dummy_request.allocateQuotaRequest.allocateOperation.operationId))

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    def test_should_not_send_the_request_if_cached(self, dummy_thread_class):
        t = self._mock_transport
        self._subject.start()
        dummy_request = _make_dummy_quota_request(self.PROJECT_ID,
                                                  self.SERVICE_NAME)
        dummy_response = sc_messages.AllocateQuotaResponse(
            operationId=dummy_request.allocateQuotaRequest.allocateOperation.operationId)
        t.services.AllocateQuota.return_value = dummy_response
        expect(self._subject.allocate_quota(dummy_request)).to(equal(dummy_response))
        t.reset_mock()
        expect(self._subject.allocate_quota(dummy_request)).to(equal(dummy_response))
        expect(t.services.AllocateQuota.called).to(be_false)

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    def test_should_return_dummy_response_if_transport_fails(self, dummy_thread_class):
        self._subject.start()
        dummy_request = _make_dummy_quota_request(self.PROJECT_ID,
                                                  self.SERVICE_NAME)
        dummy_response = sc_messages.AllocateQuotaResponse(
            operationId=dummy_request.allocateQuotaRequest.allocateOperation.operationId)
        self._mock_transport.services.AllocateQuota.side_effect = exceptions.Error()
        expect(self._subject.allocate_quota(dummy_request)).to(equal(dummy_response))


class TestClientReport(unittest2.TestCase):
    SERVICE_NAME = u'report'
    PROJECT_ID = SERVICE_NAME + u'.project'

    def setUp(self):
        self._mock_transport = mock.MagicMock()
        self._subject = client.Loaders.DEFAULT.load(
            self.SERVICE_NAME,
            create_transport=lambda: self._mock_transport)

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    def test_should_start_itself_on_report_without_start(self, dummy_thread_class):
        dummy_request = _make_dummy_report_request(self.PROJECT_ID,
                                                   self.SERVICE_NAME)
        self._subject.report(dummy_request)
        assert self._subject._running

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    def test_should_not_send_the_request_if_cached(self, dummy_thread_class):
        t = self._mock_transport
        self._subject.start()
        dummy_request = _make_dummy_report_request(self.PROJECT_ID,
                                                   self.SERVICE_NAME)
        self._subject.report(dummy_request)
        expect(t.services.Report.called).to(be_false)

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    def test_should_send_a_request_if_not_cached(self, dummy_thread_class):
        self._subject = client.Loaders.NO_CACHE.load(
            self.SERVICE_NAME,
            create_transport=lambda: self._mock_transport)

        t = self._mock_transport
        self._subject.start()
        dummy_request = _make_dummy_report_request(self.PROJECT_ID,
                                                   self.SERVICE_NAME)
        self._subject.report(dummy_request)
        expect(t.services.Report.called).to(be_true)

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    def test_should_ignore_bad_transport_when_not_cached(self, dummy_thread_class):
        self._subject = client.Loaders.NO_CACHE.load(
            self.SERVICE_NAME,
            create_transport=lambda: self._mock_transport)

        self._mock_transport.services.Report.side_effect = exceptions.Error()
        self._subject.start()
        dummy_request = _make_dummy_report_request(self.PROJECT_ID,
                                                   self.SERVICE_NAME)
        self._subject.report(dummy_request)
        expect(self._mock_transport.services.Report.called).to(be_true)


class TestNoSchedulerThread(unittest2.TestCase):
    SERVICE_NAME = u'no-scheduler-thread'
    PROJECT_ID = SERVICE_NAME + u'.project'

    def setUp(self):
        self._timer = _DateTimeTimer()
        self._mock_transport = mock.MagicMock()
        self._subject = client.Loaders.DEFAULT.load(
            self.SERVICE_NAME,
            create_transport=lambda: self._mock_transport,
            timer=self._timer)
        self._no_cache_subject = client.Loaders.NO_CACHE.load(
            self.SERVICE_NAME,
            create_transport=lambda: self._mock_transport,
            timer=self._timer)

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    @mock.patch(u"endpoints_management.control.client.sched", spec=True)
    def test_should_initialize_scheduler(self, sched, thread_class):
        thread_class.return_value.start.side_effect = lambda: 1/0
        for s in (self._subject, self._no_cache_subject):
            s.start()
            expect(sched.scheduler.called).to(be_true)
            sched.reset_mock()

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    @mock.patch(u"endpoints_management.control.client.sched", spec=True)
    def test_should_not_enter_scheduler_when_there_is_no_cache(self, sched, thread_class):
        thread_class.return_value.start.side_effect = lambda: 1/0
        self._no_cache_subject.start()
        expect(sched.scheduler.called).to(be_true)
        scheduler = sched.scheduler.return_value
        expect(scheduler.enter.called).to(be_false)

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    @mock.patch(u"endpoints_management.control.client.sched", spec=True)
    def test_should_enter_scheduler_when_there_is_a_cache(self, sched, thread_class):
        thread_class.return_value.start.side_effect = lambda: 1/0
        self._subject.start()
        expect(sched.scheduler.called).to(be_true)
        scheduler = sched.scheduler.return_value
        expect(scheduler.enter.called).to(be_true)

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    @mock.patch(u"endpoints_management.control.client.sched", spec=True)
    def test_should_not_enter_scheduler_for_cached_checks(self, sched, thread_class):
        thread_class.return_value.start.side_effect = lambda: 1/0
        self._subject.start()

        # confirm scheduler is created and initialized
        expect(sched.scheduler.called).to(be_true)
        scheduler = sched.scheduler.return_value
        expect(scheduler.enter.called).to(be_true)
        scheduler.reset_mock()

        # call check once, to a cache response
        dummy_request = _make_dummy_check_request(self.PROJECT_ID,
                                                  self.SERVICE_NAME)
        dummy_response = sc_messages.CheckResponse(
            operationId=dummy_request.checkRequest.operation.operationId)
        t = self._mock_transport
        t.services.Check.return_value = dummy_response
        expect(self._subject.check(dummy_request)).to(equal(dummy_response))
        t.reset_mock()

        # call check again - response is cached...
        expect(self._subject.check(dummy_request)).to(equal(dummy_response))
        expect(self._mock_transport.services.Check.called).to(be_false)

        # ... the scheduler is not run
        expect(scheduler.run.called).to(be_false)

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    @mock.patch(u"endpoints_management.control.client.sched", spec=True)
    def test_should_enter_scheduler_for_aggregated_reports(self, sched, thread_class):
        thread_class.return_value.start.side_effect = lambda: 1/0
        self._subject.start()

        # confirm scheduler is created and initialized
        expect(sched.scheduler.called).to(be_true)
        scheduler = sched.scheduler.return_value
        expect(scheduler.enter.called).to(be_true)
        scheduler.reset_mock()

        # call report once; transport is not called, but the scheduler is run
        dummy_request = _make_dummy_report_request(self.PROJECT_ID,
                                                   self.SERVICE_NAME)
        self._subject.report(dummy_request)
        expect(self._mock_transport.services.Report.called).to(be_false)
        expect(scheduler.run.called).to(be_true)

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    def test_should_flush_report_cache_in_scheduler(self, thread_class):
        thread_class.return_value.start.side_effect = lambda: 1/0
        self._subject.start()

        # call report once; transport is not called
        dummy_request = _make_dummy_report_request(self.PROJECT_ID,
                                                   self.SERVICE_NAME)
        self._subject.report(dummy_request)  # cached a report
        expect(self._mock_transport.services.Report.called).to(be_false)
        # pass time, at least the flush interval, after which the report
        # cache to flush
        self._timer.tick()
        self._timer.tick()
        self._subject.report(dummy_request)
        expect(self._mock_transport.services.Report.called).to(be_true)

    @mock.patch(u"endpoints_management.control.client._THREAD_CLASS", spec=True)
    @mock.patch(u"endpoints_management.control.client.sched", spec=True)
    def test_should_not_run_scheduler_when_stopping(self, sched, thread_class):
        thread_class.return_value.start.side_effect = lambda: 1/0
        self._subject.start()

        # confirm scheduler is created and initialized
        expect(sched.scheduler.called).to(be_true)
        scheduler = sched.scheduler.return_value
        expect(scheduler.enter.called).to(be_true)

        # stop the subject. transport is called, but the scheduler is not run
        self._subject.report(
            _make_dummy_report_request(self.PROJECT_ID, self.SERVICE_NAME))
        scheduler.reset_mock()
        self._subject.stop()
        expect(self._mock_transport.services.Report.called).to(be_true)
        expect(scheduler.run.called).to(be_false)


class _DateTimeTimer(object):
    def __init__(self, auto=False):
        self.auto = auto
        self.time = datetime.datetime.utcfromtimestamp(0)

    def __call__(self):
        if self.auto:
            self.tick()
        return self.time

    def tick(self):
        self.time += datetime.timedelta(seconds=1)



# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from dogpile import cache

from . import suppliers, tokens


cache.register_backend("lru_cache", "endpoints_management.auth.caches", "LruBackend")


def create_authenticator(issuers_to_provider_ids, issuer_uri_configs):
    key_uri_supplier = suppliers.KeyUriSupplier(issuer_uri_configs)
    jwks_supplier = suppliers.JwksSupplier(key_uri_supplier)
    return tokens.Authenticator(issuers_to_provider_ids, jwks_supplier)



# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""distribution provides funcs for working with `Distribution` instances.

:func:`create_exponential`, :func:`create_linear`, :func:`create_linear`
construct new `Distribution` instances initialized with different types
of buckets a `Distribution` can have. They are factory functions that
include assertions that make sure that the Distribution instances are
in the correct state.

:func:`add_sample` adds a sample to an existing distribution instance

:func:`merge` merges two distribution instances

"""

from __future__ import absolute_import
from __future__ import division

import bisect
import logging
import math

from . import sc_messages

_logger = logging.getLogger(__name__)


_BAD_NUM_FINITE_BUCKETS = u'number of finite buckets should be > 0'
_BAD_FLOAT_ARG = u'%s should be > %f'


def create_exponential(num_finite_buckets, growth_factor, scale):
    """Creates a new instance of distribution with exponential buckets

    Args:
       num_finite_buckets (int): initializes number of finite buckets
       growth_factor (float): initializes the growth factor
       scale (float): initializes the scale

    Return:
       :class:`endpoints_management.gen.servicecontrol_v1_messages.Distribution`

    Raises:
       ValueError: if the args are invalid for creating an instance
    """
    if num_finite_buckets <= 0:
        raise ValueError(_BAD_NUM_FINITE_BUCKETS)
    if growth_factor <= 1.0:
        raise ValueError(_BAD_FLOAT_ARG % (u'growth factor', 1.0))
    if scale <= 0.0:
        raise ValueError(_BAD_FLOAT_ARG % (u'scale', 0.0))
    return sc_messages.Distribution(
        bucketCounts=[0] * (num_finite_buckets + 2),
        exponentialBuckets=sc_messages.ExponentialBuckets(
            numFiniteBuckets=num_finite_buckets,
            growthFactor=growth_factor,
            scale=scale))


def create_linear(num_finite_buckets, width, offset):
    """Creates a new instance of distribution with linear buckets.

    Args:
       num_finite_buckets (int): initializes number of finite buckets
       width (float): initializes the width of each bucket
       offset (float): initializes the offset

    Return:
       :class:`endpoints_management.gen.servicecontrol_v1_messages.Distribution`

    Raises:
       ValueError: if the args are invalid for creating an instance
    """
    if num_finite_buckets <= 0:
        raise ValueError(_BAD_NUM_FINITE_BUCKETS)
    if width <= 0.0:
        raise ValueError(_BAD_FLOAT_ARG % (u'width', 0.0))
    return sc_messages.Distribution(
        bucketCounts=[0] * (num_finite_buckets + 2),
        linearBuckets=sc_messages.LinearBuckets(
            numFiniteBuckets=num_finite_buckets,
            width=width,
            offset=offset))


def create_explicit(bounds):
    """Creates a new instance of distribution with explicit buckets.

    bounds is an iterable of ordered floats that define the explicit buckets

    Args:
       bounds (iterable[float]): initializes the bounds

    Return:
       :class:`endpoints_management.gen.servicecontrol_v1_messages.Distribution`

    Raises:
       ValueError: if the args are invalid for creating an instance
    """
    safe_bounds = sorted(float(x) for x in bounds)
    if len(safe_bounds) != len(set(safe_bounds)):
        raise ValueError(u'Detected two elements of bounds that are the same')
    return sc_messages.Distribution(
        bucketCounts=[0] * (len(safe_bounds) + 1),
        explicitBuckets=sc_messages.ExplicitBuckets(bounds=safe_bounds))


def add_sample(a_float, dist):
    """Adds `a_float` to `dist`, updating its existing buckets.

    Args:
      a_float (float): a new value
      dist (:class:`endpoints_management.gen.servicecontrol_v1_messages.Distribution`):
        the Distribution being updated

    Raises:
      ValueError: if `dist` does not have known bucket options defined
      ValueError: if there are not enough bucket count fields in `dist`
    """
    dist_type, _ = _detect_bucket_option(dist)
    if dist_type == u'exponentialBuckets':
        _update_general_statistics(a_float, dist)
        _update_exponential_bucket_count(a_float, dist)
    elif dist_type == u'linearBuckets':
        _update_general_statistics(a_float, dist)
        _update_linear_bucket_count(a_float, dist)
    elif dist_type == u'explicitBuckets':
        _update_general_statistics(a_float, dist)
        _update_explicit_bucket_count(a_float, dist)
    else:
        _logger.error(u'Could not determine bucket option type for %s', dist)
        raise ValueError(u'Unknown bucket option type')


def merge(prior, latest):
    """Merge `prior` into `latest`.

    N.B, this mutates latest. It ensures that the statistics and histogram are
    updated to correctly include the original values from both instances.

    Args:
      prior (:class:`endpoints_management.gen.servicecontrol_v1_messages.Distribution`):
        an instance
      latest (:class:`endpoints_management.gen.servicecontrol_v1_messages.Distribution`):
        an instance to be updated

    Raises:
      ValueError: if the bucket options of `prior` and `latest` do not match
      ValueError: if the bucket counts of `prior` and `latest` do not match

    """
    if not _buckets_nearly_equal(prior, latest):
        _logger.error(u'Bucket options do not match. From %s To: %s',
                      prior,
                      latest)
        raise ValueError(u'Bucket options do not match')
    if len(prior.bucketCounts) != len(latest.bucketCounts):
        _logger.error(u'Bucket count sizes do not match. From %s To: %s',
                      prior,
                      latest)
        raise ValueError(u'Bucket count sizes do not match')
    if prior.count <= 0:
        return

    old_count = latest.count
    old_mean = latest.mean
    old_summed_variance = latest.sumOfSquaredDeviation
    bucket_counts = latest.bucketCounts

    # Update the latest
    latest.count += prior.count
    latest.maximum = max(prior.maximum, latest.maximum)
    latest.minimum = min(prior.minimum, latest.minimum)
    latest.mean = ((old_count * old_mean + prior.count * prior.mean) /
                   latest.count)
    latest.sumOfSquaredDeviation = (
        old_summed_variance + prior.sumOfSquaredDeviation +
        old_count * (latest.mean - old_mean) ** 2 +
        prior.count * (latest.mean - prior.mean) ** 2)
    for i, (x, y) in enumerate(zip(prior.bucketCounts, bucket_counts)):
        bucket_counts[i] = x + y


_EPSILON = 1e-5


def _is_close_enough(x, y):
    if x is None or y is None:
        return False
    return abs(x - y) <= _EPSILON * abs(x)


# This is derived from the oneof choices of the Distribution message's
# bucket_option field in google/api/servicecontrol/v1/distribution.proto, and
# should be kept in sync with that
_DISTRIBUTION_ONEOF_FIELDS = (
    u'linearBuckets', u'exponentialBuckets', u'explicitBuckets')


def _detect_bucket_option(distribution):
    for f in _DISTRIBUTION_ONEOF_FIELDS:
        value = distribution.get_assigned_value(f)
        if value is not None:
            return f, value
    return None, None


def _linear_buckets_nearly_equal(a, b):
    return ((a.numFiniteBuckets == b.numFiniteBuckets) and
            _is_close_enough(a.width, b.width) or
            _is_close_enough(a.offset, b.offset))


def _exponential_buckets_nearly_equal(a, b):
    return ((a.numFiniteBuckets == b.numFiniteBuckets) and
            _is_close_enough(a.growthFactor, b.growthFactor) and
            _is_close_enough(a.scale, b.scale))


def _explicit_buckets_nearly_equal(a, b):
    if len(a.bounds) != len(b.bounds):
        return False
    for x, y in zip(a.bounds, b.bounds):
        if not _is_close_enough(x, y):
            return False
    return True


def _buckets_nearly_equal(a_dist, b_dist):
    """Determines whether two `Distributions` are nearly equal.

    Args:
      a_dist (:class:`Distribution`): an instance
      b_dist (:class:`Distribution`): another instance

    Return:
      boolean: `True` if the two instances are approximately equal, otherwise
        False

    """
    a_type, a_buckets = _detect_bucket_option(a_dist)
    b_type, b_buckets = _detect_bucket_option(b_dist)
    if a_type != b_type:
        return False
    elif a_type == u'linearBuckets':
        return _linear_buckets_nearly_equal(a_buckets, b_buckets)
    elif a_type == u'exponentialBuckets':
        return _exponential_buckets_nearly_equal(a_buckets, b_buckets)
    elif a_type == u'explicitBuckets':
        return _explicit_buckets_nearly_equal(a_buckets, b_buckets)
    else:
        return False


def _update_general_statistics(a_float, dist):
    """Adds a_float to distribution, updating the statistics fields.

    Args:
      a_float (float): a new value
      dist (:class:`endpoints_management.gen.servicecontrol_v1_messages.Distribution`):
        the Distribution being updated

    """
    if not dist.count:
        dist.count = 1
        dist.maximum = a_float
        dist.minimum = a_float
        dist.mean = a_float
        dist.sumOfSquaredDeviation = 0
    else:
        old_count = dist.count
        old_mean = dist.mean
        new_mean = ((old_count * old_mean) + a_float) / (old_count + 1)
        delta_sum_squares = (a_float - old_mean) * (a_float - new_mean)
        dist.count += 1
        dist.mean = new_mean
        dist.maximum = max(a_float, dist.maximum)
        dist.minimum = min(a_float, dist.minimum)
        dist.sumOfSquaredDeviation += delta_sum_squares


_BAD_UNSET_BUCKETS = u'cannot update a distribution with unset %s'
_BAD_LOW_BUCKET_COUNT = u'cannot update a distribution with a low bucket count'


def _update_exponential_bucket_count(a_float, dist):
    """Adds `a_float` to `dist`, updating its exponential buckets.

    Args:
      a_float (float): a new value
      dist (:class:`endpoints_management.gen.servicecontrol_v1_messages.Distribution`):
        the Distribution being updated

    Raises:
      ValueError: if `dist` does not already have exponential buckets defined
      ValueError: if there are not enough bucket count fields in `dist`
    """
    buckets = dist.exponentialBuckets
    if buckets is None:
        raise ValueError(_BAD_UNSET_BUCKETS % (u'exponential buckets'))
    bucket_counts = dist.bucketCounts
    num_finite_buckets = buckets.numFiniteBuckets
    if len(bucket_counts) < num_finite_buckets + 2:
        raise ValueError(_BAD_LOW_BUCKET_COUNT)
    scale = buckets.scale
    factor = buckets.growthFactor
    if (a_float <= scale):
        index = 0
    else:
        index = 1 + int((math.log(a_float / scale) / math.log(factor)))
        index = min(index, num_finite_buckets + 1)
    bucket_counts[index] += 1
    _logger.debug(u'scale:%f, factor:%f, sample:%f, index:%d',
                  scale, factor, a_float, index)


def _update_linear_bucket_count(a_float, dist):
    """Adds `a_float` to `dist`, updating the its linear buckets.

    Args:
      a_float (float): a new value
      dist (:class:`endpoints_management.gen.servicecontrol_v1_messages.Distribution`):
        the Distribution being updated

    Raises:
      ValueError: if `dist` does not already have linear buckets defined
      ValueError: if there are not enough bucket count fields in `dist`
    """
    buckets = dist.linearBuckets
    if buckets is None:
        raise ValueError(_BAD_UNSET_BUCKETS % (u'linear buckets'))
    bucket_counts = dist.bucketCounts
    num_finite_buckets = buckets.numFiniteBuckets
    if len(bucket_counts) < num_finite_buckets + 2:
        raise ValueError(_BAD_LOW_BUCKET_COUNT)
    width = buckets.width
    lower = buckets.offset
    upper = lower + (num_finite_buckets * width)
    if a_float < lower:
        index = 0
    elif a_float >= upper:
        index = num_finite_buckets + 1
    else:
        index = 1 + int(((a_float - lower) / width))
    bucket_counts[index] += 1
    _logger.debug(u'upper:%f, lower:%f, width:%f, sample:%f, index:%d',
                  upper, lower, width, a_float, index)


def _update_explicit_bucket_count(a_float, dist):
    """Adds `a_float` to `dist`, updating its explicit buckets.

    Args:
      a_float (float): a new value
      dist (:class:`endpoints_management.gen.servicecontrol_v1_messages.Distribution`):
        the Distribution being updated

    Raises:
      ValueError: if `dist` does not already have explict buckets defined
      ValueError: if there are not enough bucket count fields in `dist`
    """
    buckets = dist.explicitBuckets
    if buckets is None:
        raise ValueError(_BAD_UNSET_BUCKETS % (u'explicit buckets'))
    bucket_counts = dist.bucketCounts
    bounds = buckets.bounds
    if len(bucket_counts) < len(bounds) + 1:
        raise ValueError(_BAD_LOW_BUCKET_COUNT)
    bucket_counts[bisect.bisect(bounds, a_float)] += 1



# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generated message classes for servicecontrol version v1.

Google Service Control provides control plane functionality to managed
services, such as logging, monitoring, and status checks.
"""
# NOTE: This file is autogenerated and should not be edited by hand.

from apitools.base.protorpclite import messages as _messages
from apitools.base.py import encoding
from apitools.base.py import extra_types


package = 'servicecontrol'


class AllocateQuotaRequest(_messages.Message):
  """Request message for the AllocateQuota method.

  Enums:
    AllocationModeValueValuesEnum: Allocation mode for this operation.
      Deprecated: use QuotaMode inside the QuotaOperation.

  Fields:
    allocateOperation: Operation that describes the quota allocation.
    allocationMode: Allocation mode for this operation. Deprecated: use
      QuotaMode inside the QuotaOperation.
    serviceConfigId: Specifies which version of service configuration should
      be used to process the request. If unspecified or no matching version
      can be found, the latest one will be used.
  """

  class AllocationModeValueValuesEnum(_messages.Enum):
    """Allocation mode for this operation. Deprecated: use QuotaMode inside
    the QuotaOperation.

    Values:
      UNSPECIFIED: <no description>
      NORMAL: Allocates quota for the amount specified in the service
        configuration or specified using the quota_metrics. If the amount is
        higher than the available quota, allocation error will be returned and
        no quota will be allocated.
      BEST_EFFORT: Allocates quota for the amount specified in the service
        configuration or specified using the quota_metrics. If the amount is
        higher than the available quota, request does not fail but all
        available quota will be allocated.
      CHECK_ONLY: Only checks if there is enough quota available and does not
        change the available quota. No lock is placed on the available quota
        either.
    """
    UNSPECIFIED = 0
    NORMAL = 1
    BEST_EFFORT = 2
    CHECK_ONLY = 3

  allocateOperation = _messages.MessageField('QuotaOperation', 1)
  allocationMode = _messages.EnumField('AllocationModeValueValuesEnum', 2)
  serviceConfigId = _messages.StringField(3)


class AllocateQuotaResponse(_messages.Message):
  """Response message for the AllocateQuota method.

  Fields:
    allocateErrors: Indicates the decision of the allocate.
    operationId: The same operation_id value used in the AllocateQuotaRequest.
      Used for logging and diagnostics purposes.
    quotaMetrics: Quota metrics to indicate the result of allocation.
      Depending on the request, one or more of the following metrics will be
      included:  1. For rate quota, per quota group or per quota metric
      incremental usage will be specified using the following delta metric:
      "serviceruntime.googleapis.com/api/consumer/quota_used_count"  2. For
      allocation quota, per quota metric total usage will be specified using
      the following gauge metric:
      "serviceruntime.googleapis.com/allocation/consumer/quota_used_count"  3.
      For both rate quota and allocation quota, the quota limit reached
      condition will be specified using the following boolean metric:
      "serviceruntime.googleapis.com/quota/exceeded"  4. For allocation quota,
      value for each quota limit associated with the metrics will be specified
      using the following gauge metric:
      "serviceruntime.googleapis.com/quota/limit"
    serviceConfigId: ID of the actual config used to process the request.
  """

  allocateErrors = _messages.MessageField('QuotaError', 1, repeated=True)
  operationId = _messages.StringField(2)
  quotaMetrics = _messages.MessageField('MetricValueSet', 3, repeated=True)
  serviceConfigId = _messages.StringField(4)


class AuditLog(_messages.Message):
  """Common audit log format for Google Cloud Platform API operations.

  Messages:
    RequestValue: The operation request. This may not include all request
      parameters, such as those that are too large, privacy-sensitive, or
      duplicated elsewhere in the log record. It should never include user-
      generated data, such as file contents. When the JSON object represented
      here has a proto equivalent, the proto name will be indicated in the
      `@type` property.
    ResponseValue: The operation response. This may not include all response
      elements, such as those that are too large, privacy-sensitive, or
      duplicated elsewhere in the log record. It should never include user-
      generated data, such as file contents. When the JSON object represented
      here has a proto equivalent, the proto name will be indicated in the
      `@type` property.
    ServiceDataValue: Other service-specific data about the request, response,
      and other activities.

  Fields:
    authenticationInfo: Authentication information.
    authorizationInfo: Authorization information. If there are multiple
      resources or permissions involved, then there is one AuthorizationInfo
      element for each {resource, permission} tuple.
    methodName: The name of the service method or operation. For API calls,
      this should be the name of the API method. For example,
      "google.datastore.v1.Datastore.RunQuery"
      "google.logging.v1.LoggingService.DeleteLog"
    numResponseItems: The number of items returned from a List or Query API
      method, if applicable.
    request: The operation request. This may not include all request
      parameters, such as those that are too large, privacy-sensitive, or
      duplicated elsewhere in the log record. It should never include user-
      generated data, such as file contents. When the JSON object represented
      here has a proto equivalent, the proto name will be indicated in the
      `@type` property.
    requestMetadata: Metadata about the operation.
    resourceName: The resource or collection that is the target of the
      operation. The name is a scheme-less URI, not including the API service
      name. For example:      "shelves/SHELF_ID/books"
      "shelves/SHELF_ID/books/BOOK_ID"
    response: The operation response. This may not include all response
      elements, such as those that are too large, privacy-sensitive, or
      duplicated elsewhere in the log record. It should never include user-
      generated data, such as file contents. When the JSON object represented
      here has a proto equivalent, the proto name will be indicated in the
      `@type` property.
    serviceData: Other service-specific data about the request, response, and
      other activities.
    serviceName: The name of the API service performing the operation. For
      example, `"datastore.googleapis.com"`.
    status: The status of the overall operation.
  """

  @encoding.MapUnrecognizedFields('additionalProperties')
  class RequestValue(_messages.Message):
    """The operation request. This may not include all request parameters,
    such as those that are too large, privacy-sensitive, or duplicated
    elsewhere in the log record. It should never include user-generated data,
    such as file contents. When the JSON object represented here has a proto
    equivalent, the proto name will be indicated in the `@type` property.

    Messages:
      AdditionalProperty: An additional property for a RequestValue object.

    Fields:
      additionalProperties: Properties of the object.
    """

    class AdditionalProperty(_messages.Message):
      """An additional property for a RequestValue object.

      Fields:
        key: Name of the additional property.
        value: A extra_types.JsonValue attribute.
      """

      key = _messages.StringField(1)
      value = _messages.MessageField('extra_types.JsonValue', 2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  @encoding.MapUnrecognizedFields('additionalProperties')
  class ResponseValue(_messages.Message):
    """The operation response. This may not include all response elements,
    such as those that are too large, privacy-sensitive, or duplicated
    elsewhere in the log record. It should never include user-generated data,
    such as file contents. When the JSON object represented here has a proto
    equivalent, the proto name will be indicated in the `@type` property.

    Messages:
      AdditionalProperty: An additional property for a ResponseValue object.

    Fields:
      additionalProperties: Properties of the object.
    """

    class AdditionalProperty(_messages.Message):
      """An additional property for a ResponseValue object.

      Fields:
        key: Name of the additional property.
        value: A extra_types.JsonValue attribute.
      """

      key = _messages.StringField(1)
      value = _messages.MessageField('extra_types.JsonValue', 2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  @encoding.MapUnrecognizedFields('additionalProperties')
  class ServiceDataValue(_messages.Message):
    """Other service-specific data about the request, response, and other
    activities.

    Messages:
      AdditionalProperty: An additional property for a ServiceDataValue
        object.

    Fields:
      additionalProperties: Properties of the object. Contains field @type
        with type URL.
    """

    class AdditionalProperty(_messages.Message):
      """An additional property for a ServiceDataValue object.

      Fields:
        key: Name of the additional property.
        value: A extra_types.JsonValue attribute.
      """

      key = _messages.StringField(1)
      value = _messages.MessageField('extra_types.JsonValue', 2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  authenticationInfo = _messages.MessageField('AuthenticationInfo', 1)
  authorizationInfo = _messages.MessageField('AuthorizationInfo', 2, repeated=True)
  methodName = _messages.StringField(3)
  numResponseItems = _messages.IntegerField(4)
  request = _messages.MessageField('RequestValue', 5)
  requestMetadata = _messages.MessageField('RequestMetadata', 6)
  resourceName = _messages.StringField(7)
  response = _messages.MessageField('ResponseValue', 8)
  serviceData = _messages.MessageField('ServiceDataValue', 9)
  serviceName = _messages.StringField(10)
  status = _messages.MessageField('Status', 11)


class AuthenticationInfo(_messages.Message):
  """Authentication information for the operation.

  Messages:
    ThirdPartyPrincipalValue: The third party identification (if any) of the
      authenticated user making the request. When the JSON object represented
      here has a proto equivalent, the proto name will be indicated in the
      `@type` property.

  Fields:
    authoritySelector: The authority selector specified by the requestor, if
      any. It is not guaranteed that the principal was allowed to use this
      authority.
    principalEmail: The email address of the authenticated user (or service
      account on behalf of third party principal) making the request. For
      privacy reasons, the principal email address is redacted for all read-
      only operations that fail with a "permission denied" error.
    thirdPartyPrincipal: The third party identification (if any) of the
      authenticated user making the request. When the JSON object represented
      here has a proto equivalent, the proto name will be indicated in the
      `@type` property.
  """

  @encoding.MapUnrecognizedFields('additionalProperties')
  class ThirdPartyPrincipalValue(_messages.Message):
    """The third party identification (if any) of the authenticated user
    making the request. When the JSON object represented here has a proto
    equivalent, the proto name will be indicated in the `@type` property.

    Messages:
      AdditionalProperty: An additional property for a
        ThirdPartyPrincipalValue object.

    Fields:
      additionalProperties: Properties of the object.
    """

    class AdditionalProperty(_messages.Message):
      """An additional property for a ThirdPartyPrincipalValue object.

      Fields:
        key: Name of the additional property.
        value: A extra_types.JsonValue attribute.
      """

      key = _messages.StringField(1)
      value = _messages.MessageField('extra_types.JsonValue', 2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  authoritySelector = _messages.StringField(1)
  principalEmail = _messages.StringField(2)
  thirdPartyPrincipal = _messages.MessageField('ThirdPartyPrincipalValue', 3)


class AuthorizationInfo(_messages.Message):
  """Authorization information for the operation.

  Fields:
    granted: Whether or not authorization for `resource` and `permission` was
      granted.
    permission: The required IAM permission.
    resource: The resource being accessed, as a REST-style string. For
      example:
      bigquery.googleapis.com/projects/PROJECTID/datasets/DATASETID
  """

  granted = _messages.BooleanField(1)
  permission = _messages.StringField(2)
  resource = _messages.StringField(3)


class CheckError(_messages.Message):
  """Defines the errors to be returned in
  google.api.servicecontrol.v1.CheckResponse.check_errors.

  Enums:
    CodeValueValuesEnum: The error code.

  Fields:
    code: The error code.
    detail: Free-form text providing details on the error cause of the error.
  """

  class CodeValueValuesEnum(_messages.Enum):
    """The error code.

    Values:
      ERROR_CODE_UNSPECIFIED: This is never used in `CheckResponse`.
      NOT_FOUND: The consumer's project id was not found. Same as
        google.rpc.Code.NOT_FOUND.
      PERMISSION_DENIED: The consumer doesn't have access to the specified
        resource. Same as google.rpc.Code.PERMISSION_DENIED.
      RESOURCE_EXHAUSTED: Quota check failed. Same as
        google.rpc.Code.RESOURCE_EXHAUSTED.
      BUDGET_EXCEEDED: Budget check failed.
      DENIAL_OF_SERVICE_DETECTED: The consumer's request has been flagged as a
        DoS attack.
      LOAD_SHEDDING: The consumer's request should be rejected in order to
        protect the service from being overloaded.
      ABUSER_DETECTED: The consumer has been flagged as an abuser.
      SERVICE_NOT_ACTIVATED: The consumer hasn't activated the service.
      VISIBILITY_DENIED: The consumer cannot access the service due to
        visibility configuration.
      BILLING_DISABLED: The consumer cannot access the service because billing
        is disabled.
      PROJECT_DELETED: The consumer's project has been marked as deleted (soft
        deletion).
      PROJECT_INVALID: The consumer's project number or id does not represent
        a valid project.
      IP_ADDRESS_BLOCKED: The IP address of the consumer is invalid for the
        specific consumer project.
      REFERER_BLOCKED: The referer address of the consumer request is invalid
        for the specific consumer project.
      CLIENT_APP_BLOCKED: The client application of the consumer request is
        invalid for the specific consumer project.
      API_TARGET_BLOCKED: The API targeted by this request is invalid for the
        specified consumer project.
      API_KEY_INVALID: The consumer's API key is invalid.
      API_KEY_EXPIRED: The consumer's API Key has expired.
      API_KEY_NOT_FOUND: The consumer's API Key was not found in config
        record.
      SPATULA_HEADER_INVALID: The consumer's spatula header is invalid.
      LOAS_ROLE_INVALID: The consumer's LOAS role is invalid.
      NO_LOAS_PROJECT: The consumer's LOAS role has no associated project.
      LOAS_PROJECT_DISABLED: The consumer's LOAS project is not `ACTIVE` in
        LoquatV2.
      SECURITY_POLICY_VIOLATED: Request is not allowed as per security
        policies defined in Org Policy.
      NAMESPACE_LOOKUP_UNAVAILABLE: The backend server for looking up project
        id/number is unavailable.
      SERVICE_STATUS_UNAVAILABLE: The backend server for checking service
        status is unavailable.
      BILLING_STATUS_UNAVAILABLE: The backend server for checking billing
        status is unavailable.
      QUOTA_CHECK_UNAVAILABLE: The backend server for checking quota limits is
        unavailable.
      LOAS_PROJECT_LOOKUP_UNAVAILABLE: The Spanner for looking up LOAS project
        is unavailable.
      CLOUD_RESOURCE_MANAGER_BACKEND_UNAVAILABLE: Cloud Resource Manager
        backend server is unavailable.
      SECURITY_POLICY_BACKEND_UNAVAILABLE: Backend server for evaluating
        security policy is unavailable.
    """
    ERROR_CODE_UNSPECIFIED = 0
    NOT_FOUND = 1
    PERMISSION_DENIED = 2
    RESOURCE_EXHAUSTED = 3
    BUDGET_EXCEEDED = 4
    DENIAL_OF_SERVICE_DETECTED = 5
    LOAD_SHEDDING = 6
    ABUSER_DETECTED = 7
    SERVICE_NOT_ACTIVATED = 8
    VISIBILITY_DENIED = 9
    BILLING_DISABLED = 10
    PROJECT_DELETED = 11
    PROJECT_INVALID = 12
    IP_ADDRESS_BLOCKED = 13
    REFERER_BLOCKED = 14
    CLIENT_APP_BLOCKED = 15
    API_TARGET_BLOCKED = 16
    API_KEY_INVALID = 17
    API_KEY_EXPIRED = 18
    API_KEY_NOT_FOUND = 19
    SPATULA_HEADER_INVALID = 20
    LOAS_ROLE_INVALID = 21
    NO_LOAS_PROJECT = 22
    LOAS_PROJECT_DISABLED = 23
    SECURITY_POLICY_VIOLATED = 24
    NAMESPACE_LOOKUP_UNAVAILABLE = 25
    SERVICE_STATUS_UNAVAILABLE = 26
    BILLING_STATUS_UNAVAILABLE = 27
    QUOTA_CHECK_UNAVAILABLE = 28
    LOAS_PROJECT_LOOKUP_UNAVAILABLE = 29
    CLOUD_RESOURCE_MANAGER_BACKEND_UNAVAILABLE = 30
    SECURITY_POLICY_BACKEND_UNAVAILABLE = 31

  code = _messages.EnumField('CodeValueValuesEnum', 1)
  detail = _messages.StringField(2)


class CheckInfo(_messages.Message):
  """A CheckInfo object.

  Fields:
    consumerInfo: Consumer info of this check.
    unusedArguments: A list of fields and label keys that are ignored by the
      server. The client doesn't need to send them for following requests to
      improve performance and allow better aggregation.
  """

  consumerInfo = _messages.MessageField('ConsumerInfo', 1)
  unusedArguments = _messages.StringField(2, repeated=True)


class CheckRequest(_messages.Message):
  """Request message for the Check method.

  Fields:
    operation: The operation to be checked.
    requestProjectSettings: Requests the project settings to be returned as
      part of the check response.
    serviceConfigId: Specifies which version of service configuration should
      be used to process the request.  If unspecified or no matching version
      can be found, the latest one will be used.
    skipActivationCheck: Indicates if service activation check should be
      skipped for this request. Default behavior is to perform the check and
      apply relevant quota.
  """

  operation = _messages.MessageField('Operation', 1)
  requestProjectSettings = _messages.BooleanField(2)
  serviceConfigId = _messages.StringField(3)
  skipActivationCheck = _messages.BooleanField(4)


class CheckResponse(_messages.Message):
  """Response message for the Check method.

  Fields:
    checkErrors: Indicate the decision of the check.  If no check errors are
      present, the service should process the operation. Otherwise the service
      should use the list of errors to determine the appropriate action.
    checkInfo: Feedback data returned from the server during processing a
      Check request.
    operationId: The same operation_id value used in the CheckRequest. Used
      for logging and diagnostics purposes.
    quotaInfo: Quota information for the check request associated with this
      response.
    serviceConfigId: The actual config id used to process the request.
  """

  checkErrors = _messages.MessageField('CheckError', 1, repeated=True)
  checkInfo = _messages.MessageField('CheckInfo', 2)
  operationId = _messages.StringField(3)
  quotaInfo = _messages.MessageField('QuotaInfo', 4)
  serviceConfigId = _messages.StringField(5)


class ConsumerInfo(_messages.Message):
  """`ConsumerInfo` provides information about the consumer project.

  Fields:
    projectNumber: The Google cloud project number, e.g. 1234567890. A value
      of 0 indicates no project number is found.
  """

  projectNumber = _messages.IntegerField(1)


class Distribution(_messages.Message):
  """Distribution represents a frequency distribution of double-valued sample
  points. It contains the size of the population of sample points plus
  additional optional information:    - the arithmetic mean of the samples   -
  the minimum and maximum of the samples   - the sum-squared-deviation of the
  samples, used to compute variance   - a histogram of the values of the
  sample points

  Fields:
    bucketCounts: The number of samples in each histogram bucket.
      `bucket_counts` are optional. If present, they must sum to the `count`
      value.  The buckets are defined below in `bucket_option`. There are N
      buckets. `bucket_counts[0]` is the number of samples in the underflow
      bucket. `bucket_counts[1]` to `bucket_counts[N-1]` are the numbers of
      samples in each of the finite buckets. And `bucket_counts[N] is the
      number of samples in the overflow bucket. See the comments of
      `bucket_option` below for more details.  Any suffix of trailing zeros
      may be omitted.
    count: The total number of samples in the distribution. Must be >= 0.
    explicitBuckets: Buckets with arbitrary user-provided width.
    exponentialBuckets: Buckets with exponentially growing width.
    linearBuckets: Buckets with constant width.
    maximum: The maximum of the population of values. Ignored if `count` is
      zero.
    mean: The arithmetic mean of the samples in the distribution. If `count`
      is zero then this field must be zero.
    minimum: The minimum of the population of values. Ignored if `count` is
      zero.
    sumOfSquaredDeviation: The sum of squared deviations from the mean:
      Sum[i=1..count]((x_i - mean)^2) where each x_i is a sample values. If
      `count` is zero then this field must be zero, otherwise validation of
      the request fails.
  """

  bucketCounts = _messages.IntegerField(1, repeated=True)
  count = _messages.IntegerField(2)
  explicitBuckets = _messages.MessageField('ExplicitBuckets', 3)
  exponentialBuckets = _messages.MessageField('ExponentialBuckets', 4)
  linearBuckets = _messages.MessageField('LinearBuckets', 5)
  maximum = _messages.FloatField(6)
  mean = _messages.FloatField(7)
  minimum = _messages.FloatField(8)
  sumOfSquaredDeviation = _messages.FloatField(9)


class EndReconciliationRequest(_messages.Message):
  """A EndReconciliationRequest object.

  Fields:
    reconciliationOperation: Operation that describes the quota
      reconciliation.
    serviceConfigId: Specifies which version of service configuration should
      be used to process the request. If unspecified or no matching version
      can be found, the latest one will be used.
  """

  reconciliationOperation = _messages.MessageField('QuotaOperation', 1)
  serviceConfigId = _messages.StringField(2)


class EndReconciliationResponse(_messages.Message):
  """A EndReconciliationResponse object.

  Fields:
    operationId: The same operation_id value used in the
      EndReconciliationRequest. Used for logging and diagnostics purposes.
    quotaMetrics: Metric values as tracked by One Platform before the
      adjustment was made. The following metrics will be included:  1. Per
      quota metric total usage will be specified using the following gauge
      metric:
      "serviceruntime.googleapis.com/allocation/consumer/quota_used_count"  2.
      Value for each quota limit associated with the metrics will be specified
      using the following gauge metric:
      "serviceruntime.googleapis.com/quota/limit"  3. Delta value of the usage
      after the reconciliation for limits associated with the metrics will be
      specified using the following metric:
      "serviceruntime.googleapis.com/allocation/reconciliation_delta" The
      delta value is defined as:   new_usage_from_client -
      existing_value_in_spanner. This metric is not defined in
      serviceruntime.yaml or in Cloud Monarch. This metric is meant for
      callers' use only. Since this metric is not defined in the monitoring
      backend, reporting on this metric will result in an error.
    reconciliationErrors: Indicates the decision of the reconciliation end.
    serviceConfigId: ID of the actual config used to process the request.
  """

  operationId = _messages.StringField(1)
  quotaMetrics = _messages.MessageField('MetricValueSet', 2, repeated=True)
  reconciliationErrors = _messages.MessageField('QuotaError', 3, repeated=True)
  serviceConfigId = _messages.StringField(4)


class ExplicitBuckets(_messages.Message):
  """Describing buckets with arbitrary user-provided width.

  Fields:
    bounds: 'bound' is a list of strictly increasing boundaries between
      buckets. Note that a list of length N-1 defines N buckets because of
      fenceposting. See comments on `bucket_options` for details.  The i'th
      finite bucket covers the interval   [bound[i-1], bound[i]) where i
      ranges from 1 to bound_size() - 1. Note that there are no finite buckets
      at all if 'bound' only contains a single element; in that special case
      the single bound defines the boundary between the underflow and overflow
      buckets.  bucket number                   lower bound    upper bound  i
      == 0 (underflow)              -inf           bound[i]  0 < i <
      bound_size()            bound[i-1]     bound[i]  i == bound_size()
      (overflow)    bound[i-1]     +inf
  """

  bounds = _messages.FloatField(1, repeated=True)


class ExponentialBuckets(_messages.Message):
  """Describing buckets with exponentially growing width.

  Fields:
    growthFactor: The i'th exponential bucket covers the interval   [scale *
      growth_factor^(i-1), scale * growth_factor^i) where i ranges from 1 to
      num_finite_buckets inclusive. Must be larger than 1.0.
    numFiniteBuckets: The number of finite buckets. With the underflow and
      overflow buckets, the total number of buckets is `num_finite_buckets` +
      2. See comments on `bucket_options` for details.
    scale: The i'th exponential bucket covers the interval   [scale *
      growth_factor^(i-1), scale * growth_factor^i) where i ranges from 1 to
      num_finite_buckets inclusive. Must be > 0.
  """

  growthFactor = _messages.FloatField(1)
  numFiniteBuckets = _messages.IntegerField(2, variant=_messages.Variant.INT32)
  scale = _messages.FloatField(3)


class LinearBuckets(_messages.Message):
  """Describing buckets with constant width.

  Fields:
    numFiniteBuckets: The number of finite buckets. With the underflow and
      overflow buckets, the total number of buckets is `num_finite_buckets` +
      2. See comments on `bucket_options` for details.
    offset: The i'th linear bucket covers the interval   [offset + (i-1) *
      width, offset + i * width) where i ranges from 1 to num_finite_buckets,
      inclusive.
    width: The i'th linear bucket covers the interval   [offset + (i-1) *
      width, offset + i * width) where i ranges from 1 to num_finite_buckets,
      inclusive. Must be strictly positive.
  """

  numFiniteBuckets = _messages.IntegerField(1, variant=_messages.Variant.INT32)
  offset = _messages.FloatField(2)
  width = _messages.FloatField(3)


class LogEntry(_messages.Message):
  """An individual log entry.

  Enums:
    SeverityValueValuesEnum: The severity of the log entry. The default value
      is `LogSeverity.DEFAULT`.

  Messages:
    LabelsValue: A set of user-defined (key, value) data that provides
      additional information about the log entry.
    ProtoPayloadValue: The log entry payload, represented as a protocol buffer
      that is expressed as a JSON object. The only accepted type currently is
      AuditLog.
    StructPayloadValue: The log entry payload, represented as a structure that
      is expressed as a JSON object.

  Fields:
    insertId: A unique ID for the log entry used for deduplication. If
      omitted, the implementation will generate one based on operation_id.
    labels: A set of user-defined (key, value) data that provides additional
      information about the log entry.
    name: Required. The log to which this log entry belongs. Examples:
      `"syslog"`, `"book_log"`.
    protoPayload: The log entry payload, represented as a protocol buffer that
      is expressed as a JSON object. The only accepted type currently is
      AuditLog.
    severity: The severity of the log entry. The default value is
      `LogSeverity.DEFAULT`.
    structPayload: The log entry payload, represented as a structure that is
      expressed as a JSON object.
    textPayload: The log entry payload, represented as a Unicode string
      (UTF-8).
    timestamp: The time the event described by the log entry occurred. If
      omitted, defaults to operation start time.
  """

  class SeverityValueValuesEnum(_messages.Enum):
    """The severity of the log entry. The default value is
    `LogSeverity.DEFAULT`.

    Values:
      DEFAULT: (0) The log entry has no assigned severity level.
      DEBUG: (100) Debug or trace information.
      INFO: (200) Routine information, such as ongoing status or performance.
      NOTICE: (300) Normal but significant events, such as start up, shut
        down, or a configuration change.
      WARNING: (400) Warning events might cause problems.
      ERROR: (500) Error events are likely to cause problems.
      CRITICAL: (600) Critical events cause more severe problems or outages.
      ALERT: (700) A person must take an action immediately.
      EMERGENCY: (800) One or more systems are unusable.
    """
    DEFAULT = 0
    DEBUG = 1
    INFO = 2
    NOTICE = 3
    WARNING = 4
    ERROR = 5
    CRITICAL = 6
    ALERT = 7
    EMERGENCY = 8

  @encoding.MapUnrecognizedFields('additionalProperties')
  class LabelsValue(_messages.Message):
    """A set of user-defined (key, value) data that provides additional
    information about the log entry.

    Messages:
      AdditionalProperty: An additional property for a LabelsValue object.

    Fields:
      additionalProperties: Additional properties of type LabelsValue
    """

    class AdditionalProperty(_messages.Message):
      """An additional property for a LabelsValue object.

      Fields:
        key: Name of the additional property.
        value: A string attribute.
      """

      key = _messages.StringField(1)
      value = _messages.StringField(2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  @encoding.MapUnrecognizedFields('additionalProperties')
  class ProtoPayloadValue(_messages.Message):
    """The log entry payload, represented as a protocol buffer that is
    expressed as a JSON object. The only accepted type currently is AuditLog.

    Messages:
      AdditionalProperty: An additional property for a ProtoPayloadValue
        object.

    Fields:
      additionalProperties: Properties of the object. Contains field @type
        with type URL.
    """

    class AdditionalProperty(_messages.Message):
      """An additional property for a ProtoPayloadValue object.

      Fields:
        key: Name of the additional property.
        value: A extra_types.JsonValue attribute.
      """

      key = _messages.StringField(1)
      value = _messages.MessageField('extra_types.JsonValue', 2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  @encoding.MapUnrecognizedFields('additionalProperties')
  class StructPayloadValue(_messages.Message):
    """The log entry payload, represented as a structure that is expressed as
    a JSON object.

    Messages:
      AdditionalProperty: An additional property for a StructPayloadValue
        object.

    Fields:
      additionalProperties: Properties of the object.
    """

    class AdditionalProperty(_messages.Message):
      """An additional property for a StructPayloadValue object.

      Fields:
        key: Name of the additional property.
        value: A extra_types.JsonValue attribute.
      """

      key = _messages.StringField(1)
      value = _messages.MessageField('extra_types.JsonValue', 2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  insertId = _messages.StringField(1)
  labels = _messages.MessageField('LabelsValue', 2)
  name = _messages.StringField(3)
  protoPayload = _messages.MessageField('ProtoPayloadValue', 4)
  severity = _messages.EnumField('SeverityValueValuesEnum', 5)
  structPayload = _messages.MessageField('StructPayloadValue', 6)
  textPayload = _messages.StringField(7)
  timestamp = _messages.StringField(8)


class MetricValue(_messages.Message):
  """Represents a single metric value.

  Messages:
    LabelsValue: The labels describing the metric value. See comments on
      google.api.servicecontrol.v1.Operation.labels for the overriding
      relationship.

  Fields:
    boolValue: A boolean value.
    distributionValue: A distribution value.
    doubleValue: A double precision floating point value.
    endTime: The end of the time period over which this metric value's
      measurement applies.
    int64Value: A signed 64-bit integer value.
    labels: The labels describing the metric value. See comments on
      google.api.servicecontrol.v1.Operation.labels for the overriding
      relationship.
    moneyValue: A money value.
    startTime: The start of the time period over which this metric value's
      measurement applies. The time period has different semantics for
      different metric types (cumulative, delta, and gauge). See the metric
      definition documentation in the service configuration for details.
    stringValue: A text string value.
  """

  @encoding.MapUnrecognizedFields('additionalProperties')
  class LabelsValue(_messages.Message):
    """The labels describing the metric value. See comments on
    google.api.servicecontrol.v1.Operation.labels for the overriding
    relationship.

    Messages:
      AdditionalProperty: An additional property for a LabelsValue object.

    Fields:
      additionalProperties: Additional properties of type LabelsValue
    """

    class AdditionalProperty(_messages.Message):
      """An additional property for a LabelsValue object.

      Fields:
        key: Name of the additional property.
        value: A string attribute.
      """

      key = _messages.StringField(1)
      value = _messages.StringField(2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  boolValue = _messages.BooleanField(1)
  distributionValue = _messages.MessageField('Distribution', 2)
  doubleValue = _messages.FloatField(3)
  endTime = _messages.StringField(4)
  int64Value = _messages.IntegerField(5)
  labels = _messages.MessageField('LabelsValue', 6)
  moneyValue = _messages.MessageField('Money', 7)
  startTime = _messages.StringField(8)
  stringValue = _messages.StringField(9)


class MetricValueSet(_messages.Message):
  """Represents a set of metric values in the same metric. Each metric value
  in the set should have a unique combination of start time, end time, and
  label values.

  Fields:
    metricName: The metric name defined in the service configuration.
    metricValues: The values in this metric.
  """

  metricName = _messages.StringField(1)
  metricValues = _messages.MessageField('MetricValue', 2, repeated=True)


class Money(_messages.Message):
  """Represents an amount of money with its currency type.

  Fields:
    currencyCode: The 3-letter currency code defined in ISO 4217.
    nanos: Number of nano (10^-9) units of the amount. The value must be
      between -999,999,999 and +999,999,999 inclusive. If `units` is positive,
      `nanos` must be positive or zero. If `units` is zero, `nanos` can be
      positive, zero, or negative. If `units` is negative, `nanos` must be
      negative or zero. For example $-1.75 is represented as `units`=-1 and
      `nanos`=-750,000,000.
    units: The whole units of the amount. For example if `currencyCode` is
      `"USD"`, then 1 unit is one US dollar.
  """

  currencyCode = _messages.StringField(1)
  nanos = _messages.IntegerField(2, variant=_messages.Variant.INT32)
  units = _messages.IntegerField(3)


class Operation(_messages.Message):
  """Represents information regarding an operation.

  Enums:
    ImportanceValueValuesEnum: DO NOT USE. This is an experimental field.

  Messages:
    LabelsValue: Labels describing the operation. Only the following labels
      are allowed:  - Labels describing monitored resources as defined in
      the service configuration. - Default labels of metric values. When
      specified, labels defined in the   metric value override these default.
      - The following labels defined by Google Cloud Platform:     -
      `cloud.googleapis.com/location` describing the location where the
      operation happened,     - `servicecontrol.googleapis.com/user_agent`
      describing the user agent        of the API request,     -
      `servicecontrol.googleapis.com/service_agent` describing the service
      used to handle the API request (e.g. ESP),     -
      `servicecontrol.googleapis.com/platform` describing the platform
      where the API is served (e.g. GAE, GCE, GKE).
    UserLabelsValue: User defined labels for the resource that this operation
      is associated with.

  Fields:
    consumerId: Identity of the consumer who is using the service. This field
      should be filled in for the operations initiated by a consumer, but not
      for service-initiated operations that are not related to a specific
      consumer.  This can be in one of the following formats:
      project:<project_id>,   project_number:<project_number>,
      api_key:<api_key>.
    endTime: End time of the operation. Required when the operation is used in
      ServiceController.Report, but optional when the operation is used in
      ServiceController.Check.
    importance: DO NOT USE. This is an experimental field.
    labels: Labels describing the operation. Only the following labels are
      allowed:  - Labels describing monitored resources as defined in   the
      service configuration. - Default labels of metric values. When
      specified, labels defined in the   metric value override these default.
      - The following labels defined by Google Cloud Platform:     -
      `cloud.googleapis.com/location` describing the location where the
      operation happened,     - `servicecontrol.googleapis.com/user_agent`
      describing the user agent        of the API request,     -
      `servicecontrol.googleapis.com/service_agent` describing the service
      used to handle the API request (e.g. ESP),     -
      `servicecontrol.googleapis.com/platform` describing the platform
      where the API is served (e.g. GAE, GCE, GKE).
    logEntries: Represents information to be logged.
    metricValueSets: Represents information about this operation. Each
      MetricValueSet corresponds to a metric defined in the service
      configuration. The data type used in the MetricValueSet must agree with
      the data type specified in the metric definition.  Within a single
      operation, it is not allowed to have more than one MetricValue instances
      that have the same metric names and identical label value combinations.
      If a request has such duplicated MetricValue instances, the entire
      request is rejected with an invalid argument error.
    operationId: Identity of the operation. This must be unique within the
      scope of the service that generated the operation. If the service calls
      Check() and Report() on the same operation, the two calls should carry
      the same id.  UUID version 4 is recommended, though not required. In
      scenarios where an operation is computed from existing information and
      an idempotent id is desirable for deduplication purpose, UUID version 5
      is recommended. See RFC 4122 for details.
    operationName: Fully qualified name of the operation. Reserved for future
      use.
    quotaProperties: Represents the properties needed for quota check.
      Applicable only if this operation is for a quota check request.
    resourceContainer: The resource name of the parent of a resource in the
      resource hierarchy.  This can be in one of the following formats:     -
      \u201cprojects/<project-id or project-number>\u201d     - \u201cfolders/<folder-id>\u201d
      - \u201corganizations/<organization-id>\u201d
    resourceContainers: DO NOT USE. This field is not ready for use yet.
    startTime: Required. Start time of the operation.
    userLabels: User defined labels for the resource that this operation is
      associated with.
  """

  class ImportanceValueValuesEnum(_messages.Enum):
    """DO NOT USE. This is an experimental field.

    Values:
      LOW: The API implementation may cache and aggregate the data. The data
        may be lost when rare and unexpected system failures occur.
      HIGH: The API implementation doesn't cache and aggregate the data. If
        the method returns successfully, it's guaranteed that the data has
        been persisted in durable storage.
      DEBUG: In addition to the behavior described in HIGH, DEBUG enables
        additional validation logic that is only useful during the onboarding
        process. This is only available to Google internal services and the
        service must be whitelisted by chemist-dev@google.com in order to use
        this level.
    """
    LOW = 0
    HIGH = 1
    DEBUG = 2

  @encoding.MapUnrecognizedFields('additionalProperties')
  class LabelsValue(_messages.Message):
    """Labels describing the operation. Only the following labels are allowed:
    - Labels describing monitored resources as defined in   the service
    configuration. - Default labels of metric values. When specified, labels
    defined in the   metric value override these default. - The following
    labels defined by Google Cloud Platform:     -
    `cloud.googleapis.com/location` describing the location where the
    operation happened,     - `servicecontrol.googleapis.com/user_agent`
    describing the user agent        of the API request,     -
    `servicecontrol.googleapis.com/service_agent` describing the service
    used to handle the API request (e.g. ESP),     -
    `servicecontrol.googleapis.com/platform` describing the platform
    where the API is served (e.g. GAE, GCE, GKE).

    Messages:
      AdditionalProperty: An additional property for a LabelsValue object.

    Fields:
      additionalProperties: Additional properties of type LabelsValue
    """

    class AdditionalProperty(_messages.Message):
      """An additional property for a LabelsValue object.

      Fields:
        key: Name of the additional property.
        value: A string attribute.
      """

      key = _messages.StringField(1)
      value = _messages.StringField(2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  @encoding.MapUnrecognizedFields('additionalProperties')
  class UserLabelsValue(_messages.Message):
    """User defined labels for the resource that this operation is associated
    with.

    Messages:
      AdditionalProperty: An additional property for a UserLabelsValue object.

    Fields:
      additionalProperties: Additional properties of type UserLabelsValue
    """

    class AdditionalProperty(_messages.Message):
      """An additional property for a UserLabelsValue object.

      Fields:
        key: Name of the additional property.
        value: A string attribute.
      """

      key = _messages.StringField(1)
      value = _messages.StringField(2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  consumerId = _messages.StringField(1)
  endTime = _messages.StringField(2)
  importance = _messages.EnumField('ImportanceValueValuesEnum', 3)
  labels = _messages.MessageField('LabelsValue', 4)
  logEntries = _messages.MessageField('LogEntry', 5, repeated=True)
  metricValueSets = _messages.MessageField('MetricValueSet', 6, repeated=True)
  operationId = _messages.StringField(7)
  operationName = _messages.StringField(8)
  quotaProperties = _messages.MessageField('QuotaProperties', 9)
  resourceContainer = _messages.StringField(10)
  resourceContainers = _messages.StringField(11, repeated=True)
  startTime = _messages.StringField(12)
  userLabels = _messages.MessageField('UserLabelsValue', 13)


class QuotaError(_messages.Message):
  """A QuotaError object.

  Enums:
    CodeValueValuesEnum: Error code.

  Fields:
    code: Error code.
    description: Free-form text that provides details on the cause of the
      error.
    subject: Subject to whom this error applies. See the specific enum for
      more details on this field. For example, "clientip:<ip address of
      client>" or "project:<Google developer project id>".
  """

  class CodeValueValuesEnum(_messages.Enum):
    """Error code.

    Values:
      UNSPECIFIED: This is never used.
      RESOURCE_EXHAUSTED: Quota allocation failed. Same as
        google.rpc.Code.RESOURCE_EXHAUSTED.
      OUT_OF_RANGE: Quota release failed.  This error is ONLY returned on a
        NORMAL release. More formally:  if a user requests a release of 10
        tokens, but only 5 tokens were previously allocated, in a BEST_EFFORT
        release, this will be considered a success, 5 tokens will be released,
        and the result will be "Ok".  If this is done in NORMAL mode, no
        tokens will be released, and an OUT_OF_RANGE error will be returned.
        Same as google.rpc.Code.OUT_OF_RANGE.
      BILLING_NOT_ACTIVE: Consumer cannot access the service because the
        service requires active billing.
      PROJECT_DELETED: Consumer's project has been marked as deleted (soft
        deletion).
      API_KEY_INVALID: Specified API key is invalid.
      API_KEY_EXPIRED: Specified API Key has expired.
      SPATULA_HEADER_INVALID: Consumer's spatula header is invalid.
      LOAS_ROLE_INVALID: The consumer's LOAS role is invalid.
      NO_LOAS_PROJECT: The consumer's LOAS role has no associated project.
      PROJECT_STATUS_UNAVAILABLE: The backend server for looking up project
        id/number is unavailable.
      SERVICE_STATUS_UNAVAILABLE: The backend server for checking service
        status is unavailable.
      BILLING_STATUS_UNAVAILABLE: The backend server for checking billing
        status is unavailable.
      QUOTA_SYSTEM_UNAVAILABLE: The backend server for checking quota limits
        is unavailable.
    """
    UNSPECIFIED = 0
    RESOURCE_EXHAUSTED = 1
    OUT_OF_RANGE = 2
    BILLING_NOT_ACTIVE = 3
    PROJECT_DELETED = 4
    API_KEY_INVALID = 5
    API_KEY_EXPIRED = 6
    SPATULA_HEADER_INVALID = 7
    LOAS_ROLE_INVALID = 8
    NO_LOAS_PROJECT = 9
    PROJECT_STATUS_UNAVAILABLE = 10
    SERVICE_STATUS_UNAVAILABLE = 11
    BILLING_STATUS_UNAVAILABLE = 12
    QUOTA_SYSTEM_UNAVAILABLE = 13

  code = _messages.EnumField('CodeValueValuesEnum', 1)
  description = _messages.StringField(2)
  subject = _messages.StringField(3)


class QuotaInfo(_messages.Message):
  """Contains the quota information for a quota check response.

  Messages:
    QuotaConsumedValue: Map of quota group name to the actual number of tokens
      consumed. If the quota check was not successful, then this will not be
      populated due to no quota consumption.  We are not merging this field
      with 'quota_metrics' field because of the complexity of scaling in
      Chemist client code base. For simplicity, we will keep this field for
      Castor (that scales quota usage) and 'quota_metrics' for SuperQuota
      (that doesn't scale quota usage).

  Fields:
    limitExceeded: Quota Metrics that have exceeded quota limits. For
      QuotaGroup-based quota, this is QuotaGroup.name For QuotaLimit-based
      quota, this is QuotaLimit.name See: google.api.Quota Deprecated: Use
      quota_metrics to get per quota group limit exceeded status.
    quotaConsumed: Map of quota group name to the actual number of tokens
      consumed. If the quota check was not successful, then this will not be
      populated due to no quota consumption.  We are not merging this field
      with 'quota_metrics' field because of the complexity of scaling in
      Chemist client code base. For simplicity, we will keep this field for
      Castor (that scales quota usage) and 'quota_metrics' for SuperQuota
      (that doesn't scale quota usage).
    quotaMetrics: Quota metrics to indicate the usage. Depending on the check
      request, one or more of the following metrics will be included:  1. For
      rate quota, per quota group or per quota metric incremental usage will
      be specified using the following delta metric:
      "serviceruntime.googleapis.com/api/consumer/quota_used_count"  2. For
      allocation quota, per quota metric total usage will be specified using
      the following gauge metric:
      "serviceruntime.googleapis.com/allocation/consumer/quota_used_count"  3.
      For both rate quota and allocation quota, the quota limit reached
      condition will be specified using the following boolean metric:
      "serviceruntime.googleapis.com/quota/exceeded"
  """

  @encoding.MapUnrecognizedFields('additionalProperties')
  class QuotaConsumedValue(_messages.Message):
    """Map of quota group name to the actual number of tokens consumed. If the
    quota check was not successful, then this will not be populated due to no
    quota consumption.  We are not merging this field with 'quota_metrics'
    field because of the complexity of scaling in Chemist client code base.
    For simplicity, we will keep this field for Castor (that scales quota
    usage) and 'quota_metrics' for SuperQuota (that doesn't scale quota
    usage).

    Messages:
      AdditionalProperty: An additional property for a QuotaConsumedValue
        object.

    Fields:
      additionalProperties: Additional properties of type QuotaConsumedValue
    """

    class AdditionalProperty(_messages.Message):
      """An additional property for a QuotaConsumedValue object.

      Fields:
        key: Name of the additional property.
        value: A integer attribute.
      """

      key = _messages.StringField(1)
      value = _messages.IntegerField(2, variant=_messages.Variant.INT32)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  limitExceeded = _messages.StringField(1, repeated=True)
  quotaConsumed = _messages.MessageField('QuotaConsumedValue', 2)
  quotaMetrics = _messages.MessageField('MetricValueSet', 3, repeated=True)


class QuotaOperation(_messages.Message):
  """Represents information regarding a quota operation.

  Enums:
    QuotaModeValueValuesEnum: Quota mode for this operation.

  Messages:
    LabelsValue: Labels describing the operation.

  Fields:
    consumerId: Identity of the consumer for whom this quota operation is
      being performed.  This can be in one of the following formats:
      project:<project_id>,   project_number:<project_number>,
      api_key:<api_key>.
    labels: Labels describing the operation.
    methodName: Fully qualified name of the API method for which this quota
      operation is requested. This name is used for matching quota rules or
      metric rules and billing status rules defined in service configuration.
      This field is not required if the quota operation is performed on non-
      API resources.  Example of an RPC method name:
      google.example.library.v1.LibraryService.CreateShelf
    operationId: Identity of the operation. This is expected to be unique
      within the scope of the service that generated the operation, and
      guarantees idempotency in case of retries.  UUID version 4 is
      recommended, though not required. In scenarios where an operation is
      computed from existing information and an idempotent id is desirable for
      deduplication purpose, UUID version 5 is recommended. See RFC 4122 for
      details.
    quotaMetrics: Represents information about this operation. Each
      MetricValueSet corresponds to a metric defined in the service
      configuration. The data type used in the MetricValueSet must agree with
      the data type specified in the metric definition.  Within a single
      operation, it is not allowed to have more than one MetricValue instances
      that have the same metric names and identical label value combinations.
      If a request has such duplicated MetricValue instances, the entire
      request is rejected with an invalid argument error.
    quotaMode: Quota mode for this operation.
  """

  class QuotaModeValueValuesEnum(_messages.Enum):
    """Quota mode for this operation.

    Values:
      UNSPECIFIED: <no description>
      NORMAL: For AllocateQuota request, allocates quota for the amount
        specified in the service configuration or specified using the quota
        metrics. If the amount is higher than the available quota, allocation
        error will be returned and no quota will be allocated. For
        ReleaseQuota request, this mode is supported only for precise quota
        limits. In this case, this operation releases quota for the amount
        specified in the service configuration or specified using the quota
        metrics. If the release can make used quota negative, release error
        will be returned and no quota will be released.
      BEST_EFFORT: For AllocateQuota request, this mode is supported only for
        imprecise quota limits. In this case, the operation allocates quota
        for the amount specified in the service configuration or specified
        using the quota metrics. If the amount is higher than the available
        quota, request does not fail but all available quota will be
        allocated. For ReleaseQuota request, this mode is supported for both
        precise quota limits and imprecise quota limits. In this case, this
        operation releases quota for the amount specified in the service
        configuration or specified using the quota metrics. If the release can
        make used quota negative, request does not fail but only the used
        quota will be released. After the ReleaseQuota request completes, the
        used quota will be 0, and never goes to negative.
      CHECK_ONLY: For AllocateQuota request, only checks if there is enough
        quota available and does not change the available quota. No lock is
        placed on the available quota either. Not supported for ReleaseQuota
        request.
    """
    UNSPECIFIED = 0
    NORMAL = 1
    BEST_EFFORT = 2
    CHECK_ONLY = 3

  @encoding.MapUnrecognizedFields('additionalProperties')
  class LabelsValue(_messages.Message):
    """Labels describing the operation.

    Messages:
      AdditionalProperty: An additional property for a LabelsValue object.

    Fields:
      additionalProperties: Additional properties of type LabelsValue
    """

    class AdditionalProperty(_messages.Message):
      """An additional property for a LabelsValue object.

      Fields:
        key: Name of the additional property.
        value: A string attribute.
      """

      key = _messages.StringField(1)
      value = _messages.StringField(2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  consumerId = _messages.StringField(1)
  labels = _messages.MessageField('LabelsValue', 2)
  methodName = _messages.StringField(3)
  operationId = _messages.StringField(4)
  quotaMetrics = _messages.MessageField('MetricValueSet', 5, repeated=True)
  quotaMode = _messages.EnumField('QuotaModeValueValuesEnum', 6)


class QuotaProperties(_messages.Message):
  """Represents the properties needed for quota operations.

  Enums:
    QuotaModeValueValuesEnum: Quota mode for this operation.

  Messages:
    LimitByIdsValue: LimitType IDs that should be used for checking quota. Key
      in this map should be a valid LimitType string, and the value is the ID
      to be used. For example, an entry <USER, 123> will cause all user quota
      limits to use 123 as the user ID. See google/api/quota.proto for the
      definition of LimitType. CLIENT_PROJECT: Not supported. USER: Value of
      this entry will be used for enforcing user-level quota       limits. If
      none specified, caller IP passed in the
      servicecontrol.googleapis.com/caller_ip label will be used instead.
      If the server cannot resolve a value for this LimitType, an error
      will be thrown. No validation will be performed on this ID. Deprecated:
      use servicecontrol.googleapis.com/user label to send user ID.

  Fields:
    limitByIds: LimitType IDs that should be used for checking quota. Key in
      this map should be a valid LimitType string, and the value is the ID to
      be used. For example, an entry <USER, 123> will cause all user quota
      limits to use 123 as the user ID. See google/api/quota.proto for the
      definition of LimitType. CLIENT_PROJECT: Not supported. USER: Value of
      this entry will be used for enforcing user-level quota       limits. If
      none specified, caller IP passed in the
      servicecontrol.googleapis.com/caller_ip label will be used instead.
      If the server cannot resolve a value for this LimitType, an error
      will be thrown. No validation will be performed on this ID. Deprecated:
      use servicecontrol.googleapis.com/user label to send user ID.
    quotaMode: Quota mode for this operation.
  """

  class QuotaModeValueValuesEnum(_messages.Enum):
    """Quota mode for this operation.

    Values:
      ACQUIRE: Decreases available quota by the cost specified for the
        operation. If cost is higher than available quota, operation fails and
        returns error.
      ACQUIRE_BEST_EFFORT: Decreases available quota by the cost specified for
        the operation. If cost is higher than available quota, operation does
        not fail and available quota goes down to zero but it returns error.
      CHECK: Does not change any available quota. Only checks if there is
        enough quota. No lock is placed on the checked tokens neither.
      RELEASE: Increases available quota by the operation cost specified for
        the operation.
    """
    ACQUIRE = 0
    ACQUIRE_BEST_EFFORT = 1
    CHECK = 2
    RELEASE = 3

  @encoding.MapUnrecognizedFields('additionalProperties')
  class LimitByIdsValue(_messages.Message):
    """LimitType IDs that should be used for checking quota. Key in this map
    should be a valid LimitType string, and the value is the ID to be used.
    For example, an entry <USER, 123> will cause all user quota limits to use
    123 as the user ID. See google/api/quota.proto for the definition of
    LimitType. CLIENT_PROJECT: Not supported. USER: Value of this entry will
    be used for enforcing user-level quota       limits. If none specified,
    caller IP passed in the       servicecontrol.googleapis.com/caller_ip
    label will be used instead.       If the server cannot resolve a value for
    this LimitType, an error       will be thrown. No validation will be
    performed on this ID. Deprecated: use servicecontrol.googleapis.com/user
    label to send user ID.

    Messages:
      AdditionalProperty: An additional property for a LimitByIdsValue object.

    Fields:
      additionalProperties: Additional properties of type LimitByIdsValue
    """

    class AdditionalProperty(_messages.Message):
      """An additional property for a LimitByIdsValue object.

      Fields:
        key: Name of the additional property.
        value: A string attribute.
      """

      key = _messages.StringField(1)
      value = _messages.StringField(2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  limitByIds = _messages.MessageField('LimitByIdsValue', 1)
  quotaMode = _messages.EnumField('QuotaModeValueValuesEnum', 2)


class ReleaseQuotaRequest(_messages.Message):
  """Request message for the ReleaseQuota method.

  Fields:
    releaseOperation: Operation that describes the quota release.
    serviceConfigId: Specifies which version of service configuration should
      be used to process the request. If unspecified or no matching version
      can be found, the latest one will be used.
  """

  releaseOperation = _messages.MessageField('QuotaOperation', 1)
  serviceConfigId = _messages.StringField(2)


class ReleaseQuotaResponse(_messages.Message):
  """Response message for the ReleaseQuota method.

  Fields:
    operationId: The same operation_id value used in the ReleaseQuotaRequest.
      Used for logging and diagnostics purposes.
    quotaMetrics: Quota metrics to indicate the result of release. Depending
      on the request, one or more of the following metrics will be included:
      1. For rate quota, per quota group or per quota metric released amount
      will be specified using the following delta metric:
      "serviceruntime.googleapis.com/api/consumer/quota_refund_count"  2. For
      allocation quota, per quota metric total usage will be specified using
      the following gauge metric:
      "serviceruntime.googleapis.com/allocation/consumer/quota_used_count"  3.
      For allocation quota, value for each quota limit associated with the
      metrics will be specified using the following gauge metric:
      "serviceruntime.googleapis.com/quota/limit"
    releaseErrors: Indicates the decision of the release.
    serviceConfigId: ID of the actual config used to process the request.
  """

  operationId = _messages.StringField(1)
  quotaMetrics = _messages.MessageField('MetricValueSet', 2, repeated=True)
  releaseErrors = _messages.MessageField('QuotaError', 3, repeated=True)
  serviceConfigId = _messages.StringField(4)


class ReportError(_messages.Message):
  """Represents the processing error of one Operation in the request.

  Fields:
    operationId: The Operation.operation_id value from the request.
    status: Details of the error when processing the Operation.
  """

  operationId = _messages.StringField(1)
  status = _messages.MessageField('Status', 2)


class ReportInfo(_messages.Message):
  """A ReportInfo object.

  Fields:
    operationId: The Operation.operation_id value from the request.
    quotaInfo: Quota usage info when processing the `Operation`.
  """

  operationId = _messages.StringField(1)
  quotaInfo = _messages.MessageField('QuotaInfo', 2)


class ReportRequest(_messages.Message):
  """Request message for the Report method.

  Fields:
    operations: Operations to be reported.  Typically the service should
      report one operation per request. Putting multiple operations into a
      single request is allowed, but should be used only when multiple
      operations are natually available at the time of the report.  If
      multiple operations are in a single request, the total request size
      should be no larger than 1MB. See ReportResponse.report_errors for
      partial failure behavior.
    serviceConfigId: Specifies which version of service config should be used
      to process the request.  If unspecified or no matching version can be
      found, the latest one will be used.
  """

  operations = _messages.MessageField('Operation', 1, repeated=True)
  serviceConfigId = _messages.StringField(2)


class ReportResponse(_messages.Message):
  """Response message for the Report method.

  Fields:
    reportErrors: Partial failures, one for each `Operation` in the request
      that failed processing. There are three possible combinations of the RPC
      status:  1. The combination of a successful RPC status and an empty
      `report_errors`    list indicates a complete success where all
      `Operations` in the    request are processed successfully. 2. The
      combination of a successful RPC status and a non-empty
      `report_errors` list indicates a partial success where some
      `Operations` in the request succeeded. Each    `Operation` that failed
      processing has a corresponding item    in this list. 3. A failed RPC
      status indicates a general non-deterministic failure.    When this
      happens, it's impossible to know which of the    'Operations' in the
      request succeeded or failed.
    reportInfos: Quota usage for each quota release `Operation` request.
      Fully or partially failed quota release request may or may not be
      present in `report_quota_info`. For example, a failed quota release
      request will have the current quota usage info when precise quota
      library returns the info. A deadline exceeded quota request will not
      have quota usage info.  If there is no quota release request,
      report_quota_info will be empty.
    serviceConfigId: The actual config id used to process the request.
  """

  reportErrors = _messages.MessageField('ReportError', 1, repeated=True)
  reportInfos = _messages.MessageField('ReportInfo', 2, repeated=True)
  serviceConfigId = _messages.StringField(3)


class RequestMetadata(_messages.Message):
  """Metadata about the request.

  Fields:
    callerIp: The IP address of the caller. For caller from internet, this
      will be public IPv4 or IPv6 address. For caller from GCE VM with
      external IP address, this will be the VM's external IP address. For
      caller from GCE VM without external IP address, if the VM is in the same
      GCP organization (or project) as the accessed resource, `caller_ip` will
      be the GCE VM's internal IPv4 address, otherwise it will be redacted to
      "gce-internal-ip". See https://cloud.google.com/compute/docs/vpc/ for
      more information.
    callerSuppliedUserAgent: The user agent of the caller. This information is
      not authenticated and should be treated accordingly. For example:  +
      `google-api-python-client/1.4.0`:     The request was made by the Google
      API client for Python. +   `Cloud SDK Command Line Tool apitools-
      client/1.0 gcloud/0.9.62`:     The request was made by the Google Cloud
      SDK CLI (gcloud). +   `AppEngine-Google;
      (+http://code.google.com/appengine; appid: s~my-project`:     The
      request was made from the `my-project` App Engine app. NOLINT
  """

  callerIp = _messages.StringField(1)
  callerSuppliedUserAgent = _messages.StringField(2)


class ServicecontrolServicesAllocateQuotaRequest(_messages.Message):
  """A ServicecontrolServicesAllocateQuotaRequest object.

  Fields:
    allocateQuotaRequest: A AllocateQuotaRequest resource to be passed as the
      request body.
    serviceName: Name of the service as specified in the service
      configuration. For example, `"pubsub.googleapis.com"`.  See
      google.api.Service for the definition of a service name.
  """

  allocateQuotaRequest = _messages.MessageField('AllocateQuotaRequest', 1)
  serviceName = _messages.StringField(2, required=True)


class ServicecontrolServicesCheckRequest(_messages.Message):
  """A ServicecontrolServicesCheckRequest object.

  Fields:
    checkRequest: A CheckRequest resource to be passed as the request body.
    serviceName: The service name as specified in its service configuration.
      For example, `"pubsub.googleapis.com"`.  See
      [google.api.Service](https://cloud.google.com/service-
      management/reference/rpc/google.api#google.api.Service) for the
      definition of a service name.
  """

  checkRequest = _messages.MessageField('CheckRequest', 1)
  serviceName = _messages.StringField(2, required=True)


class ServicecontrolServicesEndReconciliationRequest(_messages.Message):
  """A ServicecontrolServicesEndReconciliationRequest object.

  Fields:
    endReconciliationRequest: A EndReconciliationRequest resource to be passed
      as the request body.
    serviceName: Name of the service as specified in the service
      configuration. For example, `"pubsub.googleapis.com"`.  See
      google.api.Service for the definition of a service name.
  """

  endReconciliationRequest = _messages.MessageField('EndReconciliationRequest', 1)
  serviceName = _messages.StringField(2, required=True)


class ServicecontrolServicesReleaseQuotaRequest(_messages.Message):
  """A ServicecontrolServicesReleaseQuotaRequest object.

  Fields:
    releaseQuotaRequest: A ReleaseQuotaRequest resource to be passed as the
      request body.
    serviceName: Name of the service as specified in the service
      configuration. For example, `"pubsub.googleapis.com"`.  See
      google.api.Service for the definition of a service name.
  """

  releaseQuotaRequest = _messages.MessageField('ReleaseQuotaRequest', 1)
  serviceName = _messages.StringField(2, required=True)


class ServicecontrolServicesReportRequest(_messages.Message):
  """A ServicecontrolServicesReportRequest object.

  Fields:
    reportRequest: A ReportRequest resource to be passed as the request body.
    serviceName: The service name as specified in its service configuration.
      For example, `"pubsub.googleapis.com"`.  See
      [google.api.Service](https://cloud.google.com/service-
      management/reference/rpc/google.api#google.api.Service) for the
      definition of a service name.
  """

  reportRequest = _messages.MessageField('ReportRequest', 1)
  serviceName = _messages.StringField(2, required=True)


class ServicecontrolServicesStartReconciliationRequest(_messages.Message):
  """A ServicecontrolServicesStartReconciliationRequest object.

  Fields:
    serviceName: Name of the service as specified in the service
      configuration. For example, `"pubsub.googleapis.com"`.  See
      google.api.Service for the definition of a service name.
    startReconciliationRequest: A StartReconciliationRequest resource to be
      passed as the request body.
  """

  serviceName = _messages.StringField(1, required=True)
  startReconciliationRequest = _messages.MessageField('StartReconciliationRequest', 2)


class StandardQueryParameters(_messages.Message):
  """Query parameters accepted by all methods.

  Enums:
    FXgafvValueValuesEnum: V1 error format.
    AltValueValuesEnum: Data format for response.

  Fields:
    f__xgafv: V1 error format.
    access_token: OAuth access token.
    alt: Data format for response.
    bearer_token: OAuth bearer token.
    callback: JSONP
    fields: Selector specifying which fields to include in a partial response.
    key: API key. Your API key identifies your project and provides you with
      API access, quota, and reports. Required unless you provide an OAuth 2.0
      token.
    oauth_token: OAuth 2.0 token for the current user.
    pp: Pretty-print response.
    prettyPrint: Returns response with indentations and line breaks.
    quotaUser: Available to use for quota purposes for server-side
      applications. Can be any arbitrary string assigned to a user, but should
      not exceed 40 characters.
    trace: A tracing token of the form "token:<tokenid>" to include in api
      requests.
    uploadType: Legacy upload protocol for media (e.g. "media", "multipart").
    upload_protocol: Upload protocol for media (e.g. "raw", "multipart").
  """

  class AltValueValuesEnum(_messages.Enum):
    """Data format for response.

    Values:
      json: Responses with Content-Type of application/json
      media: Media download with context-dependent Content-Type
      proto: Responses with Content-Type of application/x-protobuf
    """
    json = 0
    media = 1
    proto = 2

  class FXgafvValueValuesEnum(_messages.Enum):
    """V1 error format.

    Values:
      _1: v1 error format
      _2: v2 error format
    """
    _1 = 0
    _2 = 1

  f__xgafv = _messages.EnumField('FXgafvValueValuesEnum', 1)
  access_token = _messages.StringField(2)
  alt = _messages.EnumField('AltValueValuesEnum', 3, default=u'json')
  bearer_token = _messages.StringField(4)
  callback = _messages.StringField(5)
  fields = _messages.StringField(6)
  key = _messages.StringField(7)
  oauth_token = _messages.StringField(8)
  pp = _messages.BooleanField(9, default=True)
  prettyPrint = _messages.BooleanField(10, default=True)
  quotaUser = _messages.StringField(11)
  trace = _messages.StringField(12)
  uploadType = _messages.StringField(13)
  upload_protocol = _messages.StringField(14)


class StartReconciliationRequest(_messages.Message):
  """A StartReconciliationRequest object.

  Fields:
    reconciliationOperation: Operation that describes the quota
      reconciliation.
    serviceConfigId: Specifies which version of service configuration should
      be used to process the request. If unspecified or no matching version
      can be found, the latest one will be used.
  """

  reconciliationOperation = _messages.MessageField('QuotaOperation', 1)
  serviceConfigId = _messages.StringField(2)


class StartReconciliationResponse(_messages.Message):
  """A StartReconciliationResponse object.

  Fields:
    operationId: The same operation_id value used in the
      StartReconciliationRequest. Used for logging and diagnostics purposes.
    quotaMetrics: Metric values as tracked by One Platform before the start of
      reconciliation. The following metrics will be included:  1. Per quota
      metric total usage will be specified using the following gauge metric:
      "serviceruntime.googleapis.com/allocation/consumer/quota_used_count"  2.
      Value for each quota limit associated with the metrics will be specified
      using the following gauge metric:
      "serviceruntime.googleapis.com/quota/limit"
    reconciliationErrors: Indicates the decision of the reconciliation start.
    serviceConfigId: ID of the actual config used to process the request.
  """

  operationId = _messages.StringField(1)
  quotaMetrics = _messages.MessageField('MetricValueSet', 2, repeated=True)
  reconciliationErrors = _messages.MessageField('QuotaError', 3, repeated=True)
  serviceConfigId = _messages.StringField(4)


class Status(_messages.Message):
  """The `Status` type defines a logical error model that is suitable for
  different programming environments, including REST APIs and RPC APIs. It is
  used by [gRPC](https://github.com/grpc). The error model is designed to be:
  - Simple to use and understand for most users - Flexible enough to meet
  unexpected needs  # Overview  The `Status` message contains three pieces of
  data: error code, error message, and error details. The error code should be
  an enum value of google.rpc.Code, but it may accept additional error codes
  if needed.  The error message should be a developer-facing English message
  that helps developers *understand* and *resolve* the error. If a localized
  user-facing error message is needed, put the localized message in the error
  details or localize it in the client. The optional error details may contain
  arbitrary information about the error. There is a predefined set of error
  detail types in the package `google.rpc` that can be used for common error
  conditions.  # Language mapping  The `Status` message is the logical
  representation of the error model, but it is not necessarily the actual wire
  format. When the `Status` message is exposed in different client libraries
  and different wire protocols, it can be mapped differently. For example, it
  will likely be mapped to some exceptions in Java, but more likely mapped to
  some error codes in C.  # Other uses  The error model and the `Status`
  message can be used in a variety of environments, either with or without
  APIs, to provide a consistent developer experience across different
  environments.  Example uses of this error model include:  - Partial errors.
  If a service needs to return partial errors to the client,     it may embed
  the `Status` in the normal response to indicate the partial     errors.  -
  Workflow errors. A typical workflow has multiple steps. Each step may
  have a `Status` message for error reporting.  - Batch operations. If a
  client uses batch request and batch response, the     `Status` message
  should be used directly inside batch response, one for     each error sub-
  response.  - Asynchronous operations. If an API call embeds asynchronous
  operation     results in its response, the status of those operations should
  be     represented directly using the `Status` message.  - Logging. If some
  API errors are stored in logs, the message `Status` could     be used
  directly after any stripping needed for security/privacy reasons.

  Messages:
    DetailsValueListEntry: A DetailsValueListEntry object.

  Fields:
    code: The status code, which should be an enum value of google.rpc.Code.
    details: A list of messages that carry the error details.  There is a
      common set of message types for APIs to use.
    message: A developer-facing error message, which should be in English. Any
      user-facing error message should be localized and sent in the
      google.rpc.Status.details field, or localized by the client.
  """

  @encoding.MapUnrecognizedFields('additionalProperties')
  class DetailsValueListEntry(_messages.Message):
    """A DetailsValueListEntry object.

    Messages:
      AdditionalProperty: An additional property for a DetailsValueListEntry
        object.

    Fields:
      additionalProperties: Properties of the object. Contains field @type
        with type URL.
    """

    class AdditionalProperty(_messages.Message):
      """An additional property for a DetailsValueListEntry object.

      Fields:
        key: Name of the additional property.
        value: A extra_types.JsonValue attribute.
      """

      key = _messages.StringField(1)
      value = _messages.MessageField('extra_types.JsonValue', 2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  code = _messages.IntegerField(1, variant=_messages.Variant.INT32)
  details = _messages.MessageField('DetailsValueListEntry', 2, repeated=True)
  message = _messages.StringField(3)


encoding.AddCustomJsonFieldMapping(
    StandardQueryParameters, 'f__xgafv', '$.xgafv')
encoding.AddCustomJsonEnumMapping(
    StandardQueryParameters.FXgafvValueValuesEnum, '_1', '1')
encoding.AddCustomJsonEnumMapping(
    StandardQueryParameters.FXgafvValueValuesEnum, '_2', '2')



# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging

from . import auth, config, control, gen

__version__ = '1.11.1'

_logger = logging.getLogger(__name__)
_logger.setLevel(logging.INFO)

USER_AGENT = u'ESP'
SERVICE_AGENT = u'EF_PYTHON/' + __version__

__all__ = ['auth', 'config', 'control', 'gen']



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import collections
import datetime
import unittest2

from expects import be, be_a, be_none, equal, expect, raise_error

from endpoints_management.control import caches, report_request


_TEST_NUM_ENTRIES = 3  # arbitrary


class TestDequeOutLRUCache(unittest2.TestCase):

    def test_constructor_should_set_up_a_default_deque(self):
        c = caches.DequeOutLRUCache(_TEST_NUM_ENTRIES)
        expect(c.out_deque).to(be_a(collections.deque))

    def test_constructor_should_fail_on_bad_deques(self):
        testf = lambda: caches.DequeOutLRUCache(_TEST_NUM_ENTRIES,
                                                out_deque=object())
        expect(testf).to(raise_error(ValueError))

    def test_constructor_should_accept_deques(self):
        a_deque = collections.deque()
        c = caches.DequeOutLRUCache(_TEST_NUM_ENTRIES, out_deque=a_deque)
        expect(c.out_deque).to(be(a_deque))

    def test_lru(self):
        lru_limit = 2
        cache = caches.DequeOutLRUCache(lru_limit)
        cache[1] = 1
        cache[2] = 2
        cache[3] = 3
        expect(len(cache)).to(equal(2))
        expect(cache[2]).to(equal(2))
        expect(cache[3]).to(equal(3))
        expect(cache.get(1)).to(be_none)
        expect(len(cache.out_deque)).to(be(1))
        cache[4] = 4
        expect(cache.get(2)).to(be_none)
        expect(len(cache.out_deque)).to(be(2))


class _Timer(object):
    def __init__(self, auto=False):
        self.auto = auto
        self.time = 0

    def __call__(self):
        if self.auto:
            self.tick()
        return self.time

    def tick(self):
        self.time += 1


_TEST_TTL = 3  # arbitrary


class TestDequeOutTTLCache(unittest2.TestCase):
    # pylint: disable=fixme
    #
    # TODO: add a ttl test based on the one in cachetools testsuite

    def test_constructor_should_set_up_a_default_deque(self):
        c = caches.DequeOutTTLCache(_TEST_NUM_ENTRIES, _TEST_TTL)
        expect(c.out_deque).to(be_a(collections.deque))

    def test_constructor_should_fail_on_bad_deques(self):
        testf = lambda: caches.DequeOutTTLCache(_TEST_NUM_ENTRIES, _TEST_TTL,
                                                out_deque=object())
        expect(testf).to(raise_error(ValueError))

    def test_constructor_should_accept_deques(self):
        a_deque = collections.deque()
        c = caches.DequeOutTTLCache(3, 3, out_deque=a_deque)
        expect(c.out_deque).to(be(a_deque))

    def test_lru(self):
        lru_limit = 2
        expiry = 100
        cache = caches.DequeOutTTLCache(lru_limit, expiry)
        cache[1] = 1
        cache[2] = 2
        cache[3] = 3
        expect(len(cache)).to(equal(2))
        expect(cache[2]).to(equal(2))
        expect(cache[3]).to(equal(3))
        expect(cache.get(1)).to(be_none)
        expect(len(cache.out_deque)).to(be(1))
        cache[4] = 4
        expect(cache.get(2)).to(be_none)
        expect(len(cache.out_deque)).to(be(2))

    def test_ttl(self):
        cache = caches.DequeOutTTLCache(2, ttl=1, timer=_Timer())
        expect(cache.timer()).to(equal(0))
        expect(cache.ttl).to(equal(1))

        cache[1] = 1
        expect(set(cache)).to(equal({1}))
        expect(len(cache)).to(equal(1))
        expect(cache[1]).to(equal(1))

        cache.timer.tick()
        expect(set(cache)).to(equal({1}))
        expect(len(cache)).to(equal(1))
        expect(cache[1]).to(equal(1))

        cache[2] = 2
        expect(set(cache)).to(equal({1, 2}))
        expect(len(cache)).to(equal(2))
        expect(cache[1]).to(equal(1))
        expect(cache[2]).to(equal(2))

        cache.timer.tick()
        expect(set(cache)).to(equal({2}))
        expect(len(cache)).to(equal(1))
        expect(cache[2]).to(equal(2))
        expect(cache.get(1)).to(be_none)


class _DateTimeTimer(object):
    def __init__(self, auto=False):
        self.auto = auto
        self.time = datetime.datetime(1970, 1, 1)

    def __call__(self):
        if self.auto:
            self.tick()
        return self.time

    def tick(self):
        self.time += datetime.timedelta(seconds=1)


class TestCreate(unittest2.TestCase):

    def test_should_fail_if_bad_options_are_used(self):
        should_fail = [
            lambda: caches.create(object()),
        ]
        for testf in should_fail:
            expect(testf).to(raise_error(ValueError))

    def test_should_return_none_if_options_is_none(self):
        expect(caches.create(None)).to(be_none)

    def test_should_return_none_if_cache_size_not_positive(self):
        should_be_none = [
            lambda: caches.create(caches.CheckOptions(num_entries=0)),
            lambda: caches.create(caches.CheckOptions(num_entries=-1)),
            lambda: caches.create(caches.ReportOptions(num_entries=0)),
            lambda: caches.create(caches.ReportOptions(num_entries=-1)),
        ]
        for testf in should_be_none:
            expect(testf()).to(be_none)

    def test_should_return_ttl_cache_if_flush_interval_is_positive(self):
        delta = datetime.timedelta(seconds=1)
        should_be_ttl = [
            lambda timer: caches.create(
                caches.CheckOptions(num_entries=1, flush_interval=delta),
                timer=timer
            ),
            lambda timer: caches.create(
                caches.ReportOptions(num_entries=1, flush_interval=delta),
                timer=timer
            ),
        ]
        for testf in should_be_ttl:
            timer = _DateTimeTimer()
            sync_cache = testf(timer)
            expect(sync_cache).to(be_a(caches.LockedObject))
            with sync_cache as cache:
                expect(cache).to(be_a(caches.DequeOutTTLCache))
                expect(cache.timer()).to(equal(0))
                cache[1] = 1
                expect(set(cache)).to(equal({1}))
                expect(cache.get(1)).to(equal(1))
                timer.tick()
                expect(cache.get(1)).to(equal(1))
                timer.tick()
                expect(cache.get(1)).to(be_none)

            # Is still TTL without the custom timer
            sync_cache = testf(None)
            expect(sync_cache).to(be_a(caches.LockedObject))
            with sync_cache as cache:
                expect(cache).to(be_a(caches.DequeOutTTLCache))

    def test_should_return_a_lru_cache_if_flush_interval_is_negative(self):
        delta = datetime.timedelta(seconds=-1)
        should_be_ttl = [
            lambda: caches.create(
                caches.CheckOptions(num_entries=1, flush_interval=delta),
            ),
            lambda: caches.create(
                caches.ReportOptions(num_entries=1, flush_interval=delta)),
        ]
        for testf in should_be_ttl:
            sync_cache = testf()
            expect(sync_cache).to(be_a(caches.LockedObject))
            with sync_cache as cache:
                expect(cache).to(be_a(caches.DequeOutLRUCache))


class TestReportOptions(unittest2.TestCase):

    def test_should_create_with_defaults(self):
        options = caches.ReportOptions()
        expect(options.num_entries).to(equal(
            caches.ReportOptions.DEFAULT_NUM_ENTRIES))
        expect(options.flush_interval).to(equal(
            caches.ReportOptions.DEFAULT_FLUSH_INTERVAL))


class TestCheckOptions(unittest2.TestCase):
    AN_INTERVAL = datetime.timedelta(milliseconds=2)
    A_LOWER_INTERVAL = datetime.timedelta(milliseconds=1)

    def test_should_create_with_defaults(self):
        options = caches.CheckOptions()
        expect(options.num_entries).to(equal(
            caches.CheckOptions.DEFAULT_NUM_ENTRIES))
        expect(options.flush_interval).to(equal(
            caches.CheckOptions.DEFAULT_FLUSH_INTERVAL))
        expect(options.expiration).to(equal(
            caches.CheckOptions.DEFAULT_EXPIRATION))

    def test_should_ignores_lower_expiration(self):
        wanted_expiration = (
            self.AN_INTERVAL + datetime.timedelta(milliseconds=1))
        options = caches.CheckOptions(flush_interval=self.AN_INTERVAL,
                                      expiration=self.A_LOWER_INTERVAL)
        expect(options.flush_interval).to(equal(self.AN_INTERVAL))
        expect(options.expiration).to(equal(wanted_expiration))
        expect(options.expiration).not_to(equal(self.A_LOWER_INTERVAL))



# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""label_descriptor provides funcs for working with `LabelDescriptor` instances.

:class:`KnownLabels` is an :class:`enum.Enum` that defines the list of known
`LabelDescriptor` instances.  Each enum instance has several fields

- label_name: the name used in the label
- kind: indicates whether the label is system or user label
- value_type: the value type used in the label
- update_label_func: a function to update the labels

"""

from __future__ import absolute_import

import base64
from enum import Enum
from . import sm_messages
from .. import USER_AGENT, SERVICE_AGENT

ValueType = sm_messages.LabelDescriptor.ValueTypeValueValuesEnum


class Kind(Enum):
    """Enumerates the known labels."""
    # pylint: disable=too-few-public-methods
    USER = 0
    SYSTEM = 1


_CANONICAL_CODES = {
    200: 0,  # OK
    400: 3,  # INVALID_ARGUMENT
    401: 16,  # UNAUTHENTICATED
    403: 7,  # PERMISSION_DENIED
    404: 5,  # NOT_FOUND
    409: 10,  # ABORTED
    412: 9,  # FAILED_PRECONDITION
    416: 11,  # OUT_OF_RANGE
    429: 8,  # RESOURCE_EXHAUSTED
    499: 1,  # CANCELLED
    500: 13,  # INTERNAL, UNKNOWN
    504: 4,  # DEADLINE_EXCEEDED
    501: 12,  # UNIMPLEMENTED
    503: 14,  # UNAVAILABLE
}


def _canonical_code(http_code):
    mapped_code = _CANONICAL_CODES.get(http_code, 0)
    if mapped_code != 0:
        return mapped_code
    elif 200 <= http_code < 300:
        return 0  # OK
    elif 400 <= http_code < 500:
        return 9  # failed precondition
    elif 500 <= http_code < 600:
        return 13  # internal
    else:
        return 2  # unknown


def set_credential_id(name, info, labels):
    # The rule to set /credential_id is:
    # 1) If api_key is available, set it as apiKey:API-KEY
    # 2) If auth issuer and audience both are available, set it as:
    #    jwtAuth:issuer=base64(issuer)&audience=base64(audience)
    if info.api_key:
        labels[name] = b'apiKey:' + info.api_key.encode('utf-8')
    elif info.auth_issuer:
        value = b'jwtAuth:issuer=' + base64.urlsafe_b64encode(info.auth_issuer.encode('utf-8'))
        if info.auth_audience:
            value += b'&audience=' + base64.urlsafe_b64encode(info.auth_audience.encode('utf-8'))
        labels[name] = value


_ERROR_TYPES = tuple(u'%dxx' % (x,) for x in range(10))


def set_error_type(name, info, labels):
    if info.response_code > 0:
        code = (info.response_code // 100) % 10
        if code < len(_ERROR_TYPES):
            labels[name] = _ERROR_TYPES[code]


def set_protocol(name, info, labels):
    labels[name] = info.protocol.name


def set_referer(name, info, labels):
    if info.referer:
        labels[name] = info.referer


def set_response_code(name, info, labels):
    labels[name] = u'%d' % (info.response_code,)


def set_response_code_class(name, info, labels):
    if info.response_code > 0:
        code = (info.response_code // 100) % 10
        if code < len(_ERROR_TYPES):
            labels[name] = _ERROR_TYPES[code]


def set_status_code(name, info, labels):
    if info.response_code > 0:
        labels[name] = u'%d' % (_canonical_code(info.response_code),)


def set_location(name, info, labels):
    if info.location:
        labels[name] = info.location


def set_api_method(name, info, labels):
    if info.api_method:
        labels[name] = info.api_method


def set_api_version(name, info, labels):
    if info.api_version:
        labels[name] = info.api_version


def set_platform(name, info, labels):
    labels[name] = info.platform.name


def set_service_agent(name, dummy_info, labels):
    labels[name] = SERVICE_AGENT


def set_user_agent(name, dummy_info, labels):
    labels[name] = USER_AGENT


def set_consumer_project(name, info, labels):
    if info.consumer_project_number > 0:
        labels[name] = unicode(info.consumer_project_number)


class KnownLabels(Enum):
    """Enumerates the known labels."""

    CREDENTIAL_ID = (
        u'/credential_id', ValueType.STRING, Kind.USER, set_credential_id)
    END_USER = (u'/end_user', ValueType.STRING, Kind.USER, None)
    END_USER_COUNTRY = (u'/end_user_country', ValueType.STRING, Kind.USER, None)
    ERROR_TYPE = (u'/error_type', ValueType.STRING, Kind.USER,
                  set_error_type)
    PROTOCOL = (u'/protocol', ValueType.STRING, Kind.USER,
                set_protocol)
    REFERER = (u'/referer', ValueType.STRING, Kind.USER,
               set_referer)
    RESPONSE_CODE = (u'/response_code', ValueType.STRING, Kind.USER,
                     set_response_code)
    RESPONSE_CODE_CLASS = (u'/response_code_class', ValueType.STRING, Kind.USER,
                           set_response_code_class)
    STATUS_CODE = (u'/status_code', ValueType.STRING, Kind.USER,
                   set_status_code)
    GAE_CLONE_ID = (
        u'appengine.googleapis.com/clone_id', ValueType.STRING, Kind.USER, None)
    GAE_MODULE_ID = (
        u'appengine.googleapis.com/module_id', ValueType.STRING, Kind.USER, None)
    GAE_REPLICA_INDEX = (
        u'appengine.googleapis.com/replica_index', ValueType.STRING, Kind.USER,
        None)
    GAE_VERSION_ID = (
        u'appengine.googleapis.com/version_id', ValueType.STRING, Kind.USER, None)
    GCP_LOCATION = (
        u'cloud.googleapis.com/location', ValueType.STRING, Kind.SYSTEM,
        set_location)
    GCP_PROJECT = (
        u'cloud.googleapis.com/project', ValueType.STRING, Kind.SYSTEM, None)
    GCP_REGION = (
        u'cloud.googleapis.com/region', ValueType.STRING, Kind.SYSTEM, None)
    GCP_RESOURCE_ID = (
        u'cloud.googleapis.com/resource_id', ValueType.STRING, Kind.USER, None)
    GCP_RESOURCE_TYPE = (
        u'cloud.googleapis.com/resource_type', ValueType.STRING, Kind.USER, None)
    GCP_SERVICE = (
        u'cloud.googleapis.com/service', ValueType.STRING, Kind.SYSTEM, None)
    GCP_ZONE = (
        u'cloud.googleapis.com/zone', ValueType.STRING, Kind.SYSTEM, None)
    GCP_UID = (
        u'cloud.googleapis.com/uid', ValueType.STRING, Kind.SYSTEM, None)
    GCP_API_METHOD = (
        u'serviceruntime.googleapis.com/api_method', ValueType.STRING, Kind.USER,
        set_api_method)
    GCP_API_VERSION = (
        u'serviceruntime.googleapis.com/api_version', ValueType.STRING, Kind.USER,
        set_api_version)
    SCC_ANDROID_CERT_FINGERPRINT = (
        'servicecontrol.googleapis.com/android_cert_fingerprint', ValueType.STRING, Kind.SYSTEM, None)
    SCC_ANDROID_PACKAGE_NAME = (
        'servicecontrol.googleapis.com/android_package_name', ValueType.STRING, Kind.SYSTEM, None)
    SCC_CALLER_IP = (
        u'servicecontrol.googleapis.com/caller_ip', ValueType.STRING, Kind.SYSTEM, None)
    SCC_IOS_BUNDLE_ID = (
        u'servicecontrol.googleapis.com/ios_bundle_id', ValueType.STRING, Kind.SYSTEM, None)
    SCC_PLATFORM = (
        u'servicecontrol.googleapis.com/platform', ValueType.STRING, Kind.SYSTEM,
        set_platform)
    SCC_REFERER = (
        u'servicecontrol.googleapis.com/referer', ValueType.STRING, Kind.SYSTEM, None)
    SCC_SERVICE_AGENT = (
        u'servicecontrol.googleapis.com/service_agent', ValueType.STRING, Kind.SYSTEM,
        set_service_agent)
    SCC_USER_AGENT = (
        u'servicecontrol.googleapis.com/user_agent', ValueType.STRING, Kind.SYSTEM,
        set_user_agent)
    SCC_CONSUMER_PROJECT = (
        u'serviceruntime.googleapis.com/consumer_project', ValueType.STRING, Kind.SYSTEM,
        set_consumer_project)

    def __init__(self, label_name, value_type, kind, update_label_func):
        """Constructor.

        update_label_func is used when updating a label in an `Operation` from a
        `ReportRequestInfo`.

        Args:
           label_name (str): the name of the label descriptor
           value_type (:class:`ValueType`): the `value type` of the described metric
           kind (:class:`Kind`): the ``kind`` of the described metric
           update_op_func (function): the func to update an operation

        """
        self.label_name = label_name
        self.kind = kind
        self.update_label_func = update_label_func
        self.value_type = value_type

    def matches(self, desc):
        """Determines if a given label descriptor matches this enum instance

        Args:
           desc (:class:`endpoints_management.gen.servicemanagement_v1_messages.LabelDescriptor`):
              the instance to test

        Return:
           `True` if desc is supported, otherwise `False`

        """
        desc_value_type = desc.valueType or ValueType.STRING  # default not parsed
        return (self.label_name == desc.key and
                self.value_type == desc_value_type)

    def do_labels_update(self, info, labels):
        """Updates a dictionary of labels using the assigned update_op_func

        Args:
           info (:class:`endpoints_management.control.report_request.Info`): the
              info instance to update
           labels (dict[string[string]]): the labels dictionary

        Return:
           `True` if desc is supported, otherwise `False`

        """
        if self.update_label_func:
            self.update_label_func(self.label_name, info, labels)

    @classmethod
    def is_supported(cls, desc):
        """Determines if the given label descriptor is supported.

        Args:
           desc (:class:`endpoints_management.gen.servicemanagement_v1_messages.LabelDescriptor`):
             the label descriptor to test

        Return:
           `True` if desc is supported, otherwise `False`

        """
        for l in cls:
            if l.matches(desc):
                return True
        return False



# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""metric_descriptor provides funcs for working with `MetricDescriptor` instances.

:class:`KnownMetrics` is an :class:`enum.Enum` that defines the list of known
`MetricDescriptor` instances.  It is a complex enumeration that includes various
attributes including

- the full metric name
- the kind of the metric
- the value type of the metric
- a func for updating :class:`Operation`s from a `ReportRequestInfo`

"""

from __future__ import absolute_import


from enum import Enum
from . import distribution, metric_value, sc_messages, MetricKind, ValueType


def _add_metric_value(name, value, an_op):
    an_op.metricValueSets.append(
        sc_messages.MetricValueSet(metricName=name, metricValues=[value]))


def _add_int64_metric_value(name, value, an_op):
    _add_metric_value(
        name, metric_value.create(int64Value=value), an_op)


def _set_int64_metric_to_constant_1(name, dummy_info, op):
    _add_int64_metric_value(name, 1, op)


def _set_int64_metric_to_constant_1_if_http_error(name, info, op):
    if info.response_code >= 400:
        _add_int64_metric_value(name, 1, op)


def _add_distribution_metric_value(name, value, an_op, distribution_args):
    d = distribution.create_exponential(*distribution_args)
    distribution.add_sample(value, d)
    _add_metric_value(
        name, metric_value.create(distributionValue=d), an_op)


_SIZE_DISTRIBUTION_ARGS = (8, 10.0, 1.0)


def _set_distribution_metric_to_request_size(name, info, an_op):
    if info.request_size >= 0:
        _add_distribution_metric_value(name, info.request_size, an_op,
                                       _SIZE_DISTRIBUTION_ARGS)


def _set_distribution_metric_to_response_size(name, info, an_op):
    if info.response_size >= 0:
        _add_distribution_metric_value(name, info.response_size, an_op,
                                       _SIZE_DISTRIBUTION_ARGS)


_TIME_DISTRIBUTION_ARGS = (8, 10.0, 1e-6)


def _set_distribution_metric_to_request_time(name, info, an_op):
    if info.request_time:
        _add_distribution_metric_value(name, info.request_time.total_seconds(),
                                       an_op, _TIME_DISTRIBUTION_ARGS)


def _set_distribution_metric_to_backend_time(name, info, an_op):
    if info.backend_time:
        _add_distribution_metric_value(name, info.backend_time.total_seconds(),
                                       an_op, _TIME_DISTRIBUTION_ARGS)


def _set_distribution_metric_to_overhead_time(name, info, an_op):
    if info.overhead_time:
        _add_distribution_metric_value(name, info.overhead_time.total_seconds(),
                                       an_op, _TIME_DISTRIBUTION_ARGS)


class Mark(Enum):
    """Enumerates the types of metric."""
    PRODUCER = 1
    CONSUMER = 2
    PRODUCER_BY_CONSUMER = 3


class KnownMetrics(Enum):
    """Enumerates the known metrics."""

    CONSUMER_REQUEST_COUNT = (
        u'serviceruntime.googleapis.com/api/consumer/request_count',
        MetricKind.DELTA,
        ValueType.INT64,
        _set_int64_metric_to_constant_1,
        Mark.CONSUMER,
    )
    PRODUCER_REQUEST_COUNT = (
        u'serviceruntime.googleapis.com/api/producer/request_count',
        MetricKind.DELTA,
        ValueType.INT64,
        _set_int64_metric_to_constant_1,
    )
    PRODUCER_BY_CONSUMER_REQUEST_COUNT = (
        u'serviceruntime.googleapis.com/api/producer/by_consumer/request_count',
        MetricKind.DELTA,
        ValueType.INT64,
        _set_int64_metric_to_constant_1,
        Mark.PRODUCER_BY_CONSUMER,
    )
    CONSUMER_REQUEST_SIZES = (
        u'serviceruntime.googleapis.com/api/consumer/request_sizes',
        MetricKind.DELTA,
        ValueType.DISTRIBUTION,
        _set_distribution_metric_to_request_size,
        Mark.CONSUMER,
    )
    PRODUCER_REQUEST_SIZES = (
        u'serviceruntime.googleapis.com/api/producer/request_sizes',
        MetricKind.DELTA,
        ValueType.DISTRIBUTION,
        _set_distribution_metric_to_request_size,
    )
    PRODUCER_BY_CONSUMER_REQUEST_SIZES = (
        'serviceruntime.googleapis.com/api/producer/by_consumer/request_sizes',
        MetricKind.DELTA,
        ValueType.DISTRIBUTION,
        _set_distribution_metric_to_request_size,
        Mark.PRODUCER_BY_CONSUMER,
    )
    CONSUMER_RESPONSE_SIZES = (
        u'serviceruntime.googleapis.com/api/consumer/response_sizes',
        MetricKind.DELTA,
        ValueType.DISTRIBUTION,
        _set_distribution_metric_to_response_size,
        Mark.CONSUMER,
    )
    PRODUCER_RESPONSE_SIZES = (
        u'serviceruntime.googleapis.com/api/producer/response_sizes',
        MetricKind.DELTA,
        ValueType.DISTRIBUTION,
        _set_distribution_metric_to_response_size,
    )
    PRODUCER_BY_CONSUMER_RESPONSE_SIZES = (
        'serviceruntime.googleapis.com/api/producer/by_consumer/response_sizes',
        MetricKind.DELTA,
        ValueType.DISTRIBUTION,
        _set_distribution_metric_to_response_size,
        Mark.PRODUCER_BY_CONSUMER,
    )
    CONSUMER_ERROR_COUNT = (
        u'serviceruntime.googleapis.com/api/consumer/error_count',
        MetricKind.DELTA,
        ValueType.INT64,
        _set_int64_metric_to_constant_1_if_http_error,
        Mark.CONSUMER,
    )
    PRODUCER_ERROR_COUNT = (
        u'serviceruntime.googleapis.com/api/producer/error_count',
        MetricKind.DELTA,
        ValueType.INT64,
        _set_int64_metric_to_constant_1_if_http_error,
    )
    PRODUCER_BY_CONSUMER_ERROR_COUNT = (
        'serviceruntime.googleapis.com/api/producer/by_consumer/error_count',
        MetricKind.DELTA,
        ValueType.INT64,
        _set_int64_metric_to_constant_1_if_http_error,
        Mark.PRODUCER_BY_CONSUMER,
    )
    CONSUMER_TOTAL_LATENCIES = (
        u'serviceruntime.googleapis.com/api/consumer/total_latencies',
        MetricKind.DELTA,
        ValueType.DISTRIBUTION,
        _set_distribution_metric_to_request_time,
        Mark.CONSUMER,
    )
    PRODUCER_TOTAL_LATENCIES = (
        u'serviceruntime.googleapis.com/api/producer/total_latencies',
        MetricKind.DELTA,
        ValueType.DISTRIBUTION,
        _set_distribution_metric_to_request_time,
    )
    PRODUCER_BY_CONSUMER_TOTAL_LATENCIES = (
        'serviceruntime.googleapis.com/api/producer/by_consumer/'
        'total_latencies',
        MetricKind.DELTA,
        ValueType.DISTRIBUTION,
        _set_distribution_metric_to_request_time,
        Mark.PRODUCER_BY_CONSUMER,
    )
    CONSUMER_BACKEND_LATENCIES = (
        u'serviceruntime.googleapis.com/api/consumer/backend_latencies',
        MetricKind.DELTA,
        ValueType.DISTRIBUTION,
        _set_distribution_metric_to_backend_time,
        Mark.CONSUMER,
    )
    PRODUCER_BACKEND_LATENCIES = (
        u'serviceruntime.googleapis.com/api/producer/backend_latencies',
        MetricKind.DELTA,
        ValueType.DISTRIBUTION,
        _set_distribution_metric_to_backend_time,
    )
    PRODUCER_BY_CONSUMER_BACKEND_LATENCIES = (
        'serviceruntime.googleapis.com/api/producer/by_consumer/'
        'backend_latencies',
        MetricKind.DELTA,
        ValueType.DISTRIBUTION,
        _set_distribution_metric_to_backend_time,
        Mark.PRODUCER_BY_CONSUMER,
    )
    CONSUMER_REQUEST_OVERHEAD_LATENCIES = (
        u'serviceruntime.googleapis.com/api/consumer/request_overhead_latencies',
        MetricKind.DELTA,
        ValueType.DISTRIBUTION,
        _set_distribution_metric_to_overhead_time,
        Mark.CONSUMER,
    )
    PRODUCER_REQUEST_OVERHEAD_LATENCIES = (
        u'serviceruntime.googleapis.com/api/producer/request_overhead_latencies',
        MetricKind.DELTA,
        ValueType.DISTRIBUTION,
        _set_distribution_metric_to_overhead_time,
    )
    PRODUCER_BY_CONSUMER_REQUEST_OVERHEAD_LATENCIES = (
        'serviceruntime.googleapis.com/api/producer/by_consumer/'
        'request_overhead_latencies',
        MetricKind.DELTA,
        ValueType.DISTRIBUTION,
        _set_distribution_metric_to_overhead_time,
        Mark.PRODUCER_BY_CONSUMER,
    )

    def __init__(self, metric_name, kind, value_type, update_op_func,
                 mark=Mark.PRODUCER):
        """Constructor.

        update_op_func is used to when updating an `Operation` from a
        `ReportRequestInfo`.

        Args:
           metric_name (str): the name of the metric descriptor
           kind (:class:`MetricKind`): the ``kind`` of the described metric
           value_type (:class:`ValueType`): the `value type` of the described metric
           update_op_func (function): the func to update an operation

        """
        self.kind = kind
        self.metric_name = metric_name
        if mark is Mark.CONSUMER:
            self.update_op_func = self._consumer_metric(update_op_func)
        elif mark is Mark.PRODUCER_BY_CONSUMER:
            self.update_op_func = self._by_consumer_metric(update_op_func)
        else:
            self.update_op_func = update_op_func
        self.value_type = value_type
        self.mark = mark

    def matches(self, desc):
        """Determines if a given metric descriptor matches this enum instance

        Args:
           desc (:class:`endpoints_management.gen.servicecontrol_v1_messages.MetricDescriptor`): the
              instance to test

        Return:
           `True` if desc is supported, otherwise `False`

        """
        return (self.metric_name == desc.name and
                self.kind == desc.metricKind and
                self.value_type == desc.valueType)

    def do_operation_update(self, info, an_op):
        """Updates an operation using the assigned update_op_func

        Args:
           info: (:class:`endpoints_management.control.report_request.Info`): the
              info instance to update
           an_op: (:class:`endpoints_management.control.report_request.Info`):
              the info instance to update

        Return:
           `True` if desc is supported, otherwise `False`

        """
        self.update_op_func(self.metric_name, info, an_op)

    def _consumer_metric(self, update_op_func):
        def resulting_updater(metric_name, info, an_op):
            if info.api_key_valid:
                update_op_func(metric_name, info, an_op)

        return resulting_updater

    def _by_consumer_metric(self, update_op_func):
        def resulting_updater(metric_name, info, an_op):
            if info.consumer_project_number > 0:
                update_op_func(metric_name, info, an_op)

        return resulting_updater

    @classmethod
    def is_supported(cls, desc):
        """Determines if the given metric descriptor is supported.

        Args:
           desc (:class:`endpoints_management.gen.servicecontrol_v1_messages.MetricDescriptor`): the
             metric descriptor to test

        Return:
           `True` if desc is supported, otherwise `False`

        """
        for m in cls:
            if m.matches(desc):
                return True
        return False
